# Proposal for new CommonFilters architecture.
#
# 100% Python, but many imports are missing, and this file will be properly split into modules soon.
# The infrastructure should now be in place for a first version.
#
#
# Remaining issues:
#
#  - Some parts of implementation still in progress ;)
#    - split into modules
#    - add all necessary imports
#    - most of the filters still need to be ported to the new architecture
#
#  - May contain a couple of bugs not caught by design-time code analysis; this will be tested and debugged
#    as soon as the implementation is in a runnable state.
#
#
#  - The current FilterManager does not allow partial resets, i.e. it only has a monolithic cleanup()
#    method that deletes *all* quads and textures.
#
#    This detracts from the cleanness of the design; see FilterStage.reconfigure() for a more detailed comment.
#
#
#  - Property factories break automatic docstring extraction (as there is no explicit "def"
#    in the derived classes).
#
#    They are, however, the cleanest way (that I could think of) to integrate the template method
#    design pattern with filter parameters. Template methods enforce OnceAndOnlyOnce for the
#    common behaviour of a "run-time parameter" or a "compile-time parameter". This makes the
#    actual filter definitions much more compact and simpler.
#
#    Properties are used to encapsulate the getter, setter and the docstring into one package,
#    which I think is cleaner than explicit getters and setters.
#
#    I have provided a mechanism to inspect filter parameters and their docstrings at runtime.
#    For now, I suggest putting complete filter descriptions into the manual (as there are for 1.8.x).
#    I can write these up if needed.
#
#
#  - Multiple instances of the same filter in the same pipeline not supported.
#
#    If we want this, we need a name mangling mechanism to make any custom shader inputs (and internal texture names)
#    unique to each instance. Also the user API needs to allow for multiple Filter instances of the same subclass.
#    All other necessary infrastructure is already in place. Maybe add this later?
#
#
#  - Currently, the custom parameter names must be made unique manually. Name mangling could generate
#    a unique identifier automatically, so that each filter instance would effectively have its own namespace
#    for the custom parameters (implemented by a naming convention).
#
#    However, name mangling requires also code mangling support in synthesize() for the individual filters,
#    which is an abstract method - to be filled in by the filter authors in derived classes.
#
#    (In addition, the custom shader input names - as given to the property factory makeRTParam() -
#     must be mangled accordingly.)
#
#    This might need another callback to retrieve the mangled name for an internal name, and would make
#    things more difficult for filter authors (but not much more so than currently with texcoord variables).
#
#    Another solution would be to play around with regexes; the rules for what makes a valid variable name in C
#    (and thus in Cg) should not be too difficult to formulate into a regex. We also know the original variable name,
#    because filters must register them.
#
#    Yet another way would be to do like the compilers do: tokenize the code returned by synthesize(),
#    replace the appropriate symbol, and regenerate the code. This is difficult, however,
#    as Python's ready-made code parsing features (such as AbstractSyntaxTree) are designed
#    to analyze Python code, not Cg.


#################################
# Notes
#################################

# Design overview:

#  The postprocessing filter system is split into three parts:
#
#     - Filter:         Abstract base class, from which actual filters are inherited. Defines the interface,
#                       providing template methods implementing common parts of filter behavior, with hooks
#                       for user-defined behavior.
#
#                       Actual filters are defined by inheriting from Filter, and implementing certain hooks
#                       (see the quick reference and examples).
#
#     - FilterPipeline: Represents the postprocessing chain. Manages quads and textures via FilterManager
#                       (the one that Panda already had in 1.8.x), and manages and owns the Filter instances.
#
#                       This is the class that the application code interacts with to request postprocessing.
#
#                       CommonFilters is a subclass of FilterPipeline, adding a backward-compatible API.
#
#     - FilterStage:    Represents a single render pass in the pipeline. FilterStages are volatile in the
#                       following sense: at each reconfigure() of the pipeline, FilterPipeline examines the
#                       enabled filters, decides how many render passes are needed, and which of the enabled
#                       filters should be assigned to which pass. It then (re-)creates and configures
#                       FilterStages.
#
#                       The main responsibility of FilterStage is the creation of the compositing shader for a
#                       single render pass. 
#
#                       FilterStage stores references to the Filter instances assigned to it, but does not own
#                       the instances.
#
#  Note that in addition to the "global" render passes represented by FilterStage instances, each Filter may
#  define its own internal stages, using several render passes internally (see e.g. BlurSharpenFilter). The
#  defining feature of Filters residing in the same FilterStage is that the input "scene textures" (managed at
#  FilterPipeline) have the same content.
#
#  The first FilterStage gets the actual scene textures from FilterManager; for any following ones, the input
#  textures are the output textures of the previous FilterStage. The last FilterStage renders into the
#  finalQuad given by FilterManager.
#
#  The purpose of the multi-pass FilterStage system is to allow for arbitrary combinations of filters, also in
#  cases where some of the filters need to "see" an intermediate version of the postprocessed image as their
#  input.


#  Filters are *owned* and *managed* by FilterPipeline, but *assigned* and *attached* to render passes.
#  Additionally, filters are *attached* also to FilterPipeline (actually, before attaching to a render pass);
# this simplifies the configuration update notification mechanism.
#
#  "Owned" means object instance ownership. In the current Filter design, each instance may have only one host
#  at a time, so each instance must belong to at most one pipeline.
#
#  "Managed" means that FilterPipeline takes care of filter instance addition/deletion in the pipeline.
#
#  "Assigned" means that a reference is stored, and that the filter instance will be included in that render
#  pass during the next reconfigure() of that render pass.
#
#  "Attached" means that the object to which the filter is attached (the "attach target") wants to be notified
#  when relevant configuration parameters in the filter change their values. The attach target uses this
#  information to reconfigure() itself automatically when needed. Also, batch update notifications (for
#  updating configuration without unnecessary intermediate recompiles) are redirected to attach targets.


#  - The system aims to be as automatic as reasonably possible. Changing the value of a filter property
#    at run-time automatically invokes the necessary magic to have the new value take effect.
#
#    Run-time parameters update the relevant shader input, while compile-time parameters trigger a recompile.
#    Recompiles are divided into three levels (see makeCTParam(), below); which level is triggered depends on
#    the kind of the property whose value has changed.
#
#    When it is known that several properties will be updated in the same call, the whole object hierarchy
#    temporarily disables recompiles until all the properties have been processed (see the class BatchUpdatable,
#    which implements this behavior).
#
#    This automation does make the framework code more complex, but it also provides considerable power
#    and ease of use.


#  - The stage abstraction. There are two separate things here: stages of the simulated image-forming process,
#    and which filters need to start a new render pass. These do not necessarily coincide.
#
#    A new render pass must begin when a filter must "see" fully up-to-date data in its input, especially
#    at pixels other than the pixel being processed. (This is because the fragment shader instances running
#    in parallel, in the same render pass, cannot communicate with each other. The up-to-date value of the
#    pixel being processed is always available even within a single render pass, but the up-to-date values
#    of other pixels are not.)
#
#    Thus, it is clear that for example, blur must begin a new render pass, because it must combine up-to-date
#    information from several pixels in its input texture. Note also that whether blur is, conceptually,
#    an optical effect or a computer-based postprocessing one, depends on interpretation.
#
#    So it would seem at first that "stage" is a property belonging to a filter type, albeit possibly a
#    user-overridable one, to allow different interpretations of what the filter represents
#    (to cover cases where multiple reasonable interpretations of the same filter are possible).
#
#    However, this is not the whole story. Consider bloom and lens flare. These are both optical effects
#    occurring in the lens - and arguably, they both want *the same* input texture: the scene as it appears
#    just before any lens-optical effects.
#
#    But even this is subject to filter implementation details: the above is the correct conclusion only
#    if the filters "in the same stage" are designed to perform "orthogonal" tasks; in this case,
#    the simulation of different lens effects that do not interact with each other.
#
#    How far such orthogonality is possible depends on what exactly the filters do; but arguably,
#    effects that need to interact belong in the same filter, anyway (to apply the simplest possible solution).
#
#    We can interpret that bloom creates isotropic glare, while the lens flare creates ghosts and the halo ring.
#    Then our complete model of lens optics is defined as the orthogonal sum of these parts. The parts
#    do not interact, and their changes to the pixel color can be combined in some appropriate way
#    (probably in this case, additive blending with one of the filters always going first).
#
#    Getting back to the filter ordering question, if blur is enabled - and it represents the camera being
#    out of focus - then bloom OR lens flare needs to begin a new render pass, after blur has been applied*,
#    but the other lens effect can then go into the same render pass. This holds regardless of the relative
#    ordering of the lens effects with respect to each other.
#
#    (* Lens-optical effects, to render correctly, must see the input image as it arrives at the
#       relevant part of the lens.)


#    The abstraction that currently provides this is a hardcoded list of available "logical stages",
#    so that filters that want to go into the same stage can simply declare the same stageName.
#    The hardcoded list defines the relative ordering of the logical stages.
#
#    Filter also has an isMergeable property, which informs FilterPipeline whether the stageName requirement
#    is absolute (indicating that a new render pass must be created for that stage), or whether the filter
#    can be merged to the end of a previous render pass (if no filter in that stage requires a new render pass).
#
#    The stage list is meant to be expanded as necessary, as new CommonFilters are added. The reason for using
#    strings instead of integers is twofold: first, it's human-readable (each stage name can be made
#    descriptive), and secondly, having an explicit list allows for inserting new values at any point in the
#    sequence without updating the filter implementations.
#
#    The merging logic in FilterPipeline allows the defined stageNames to correspond to logical stages of the
#    image-forming process, while creating only as many render passes as are actually needed (given the
#    definitions of the enabled filters). This allows for conceptually clear stage assignments in the filter
#    implementations.
#
#    Filters can be moved to a non-default stage (simply by assigning filter.stageName=..., also at run-time),
#    but currently there is no mechanism to override the list of defined stages at run-time. It would be pretty
#    trivial to add, but its implications require some thought - for example, the filter implementations expect
#    stages with the default names to exist.
#
#
# -JJ 2014-10-14


# TODO: document: textures are pipeline-level objects.
#
#       Input texture registration also takes place at pipeline level.
#
#       Thus, any property that changes which textures are needed in the compositing fshader,
#       or any property that affects texture creation (e.g. number of internal stages
#       in the filter), is conceptually, and must be implemented as, a FilterPipeline-level
#       compile-time parameter.
#
#       Currently, although conceptually it should not, this same observation affects also properties
#       that change the connections between internal stages in a filter, because those connections are set up
#       when the filter is attached to FilterStage, but FilterStage cannot destroy any existing textures
#       or quads due to the monolithic cleanup in FilterManager (hence, FilterStage must escalate reconfigure
#       into a pipeline reconfigure).
#
#       (Actually, since FilterStage-level attach() creates the internal textures and quads,
#        this means that currently, FilterStage-level attach() (for any given Filter instance)
#        can only be called once per *pipeline* reconfigure.)
#
#       This conceptual change, and (if added later) support for fine-grained cleanup in FilterManager,
#       would allow for simplifying the code by removing the delegation mechanism from FilterStage.reconfigure()
#       and letting FilterStage-level and Filter-level recompiles recompile just the changed part.

# TODO: name mangling, if added later, must mangle:
#  - variable names corresponding to the shader inputs, in user-provided shader code
#  - texture names for internal stages
#  - shader input names
#    - doesn't affect the user API; Filter.updateConfigurationFromKeywords() is still available,
#      and properties have the same names in each instance even if multiple instances exist
#    - but affects makeRTParam(), because shader input name cannot be fixed at property creation time

# TODO: check that all docstrings and comments are up to date


##################################################################
# Filter module
##################################################################

"""This module defines available stages and the Filter API.


The Filter API is used to implement new postprocessing filters.

If you are looking to use existing postprocessing filters, see the FilterPipeline class,
which provides the primary user API. Refer to the individual subclasses of Filter for
descriptions of available filters.

Both of these APIs were introduced in Panda 1.9.0. The CommonFilters class provides
backward compatibility (1.8.x) for existing scripts, working on top of FilterPipeline.

It is recommended to target new scripts to the new FilterPipeline API, which offers
more flexibility. The new API also allows the definition of custom filters that
can be mixed together in the same postprocessing pipeline with the ones provided in Panda.


Below is a quick reference for the Filter interface. See the existing filters for complete examples.

Usually, inheriting from the abstract base class Filter is the right thing to do,
but it is also legal to inherit from another existing filter, if that is more appropriate
for your particular filter.

class MyCoolEffectFilter(Filter):
    # You can use class scope constants for enums.
    #
    # MyCoolEffectFilter operating modes
    MBasic   = 0
    MAwesome = 1

    # Define __init__, calling super's init, to keep the inheritance chain free of surprises.
    #
    # (To programmers not familiar with Python's super(): if we inherit from an existing filter, and
    #  do not define __init__, Python will find the ancestor's __init__, which will likely call super's init.
    #  However, the dynamic type of the object being constructed is the inherited class, and thus its super
    #  is the ancestor itself. Thus, the ancestor's __init__ will call *itself*, causing infinite recursion.
    #  It is standard practice that super(), if used, must be used consistently.)
    #
    def __init__(self, **kwargs):
        super(self.__class__, self).__init__(**kwargs)

        # your custom code here (if needed - usually not)

    def reset(self):
        super(self.__class__, self).reset()

        # - Set self.sort and self.stageName to desired default values at this point.
        # - You can also set self.isMergeable.
        #
        # By default, the filter is mergeable to previous stages.
        #
        # Note that sort and stageName have no default, because they are specific
        # to each filter type.

        # - Set your properties to their default values. Note PROPERTIES,
        #   not their internal data members; this is needed so that resetConfiguration()
        #   (which calls reset()) does the right thing also when called for a
        #   running filter instance.

        self.sort = 10
        self.param1 = 0.75
        self.param2 = 3.14
        self.param3 = MyCoolEffectFilter.MAwesome

    def setup(self, phase):
        if phase == "pipeline":
            # Register input textures, custom shader inputs if any, and aux bits if any.
            # See existing filters for examples.
            #
            # Also register the filter as updatable if you need to provide update().
            #
        elif phase == "stage":
            # Create internal stages here, if needed (see BlurSharpenFilter for an example).

    def setdown(self, phase):
        # If you allocate something really exotic in setup(), this is provided to tear down those resources.
        #
        # The most common resources used by internal stages are cleaned up automatically by detach();
        # it is very rare that there is a need to define setdown().
        #
        # This method is part of cleanup; it must not raise exceptions. (Cleanup may be triggered
        # in response to exceptions.)

    def make(self):
        # Load or compile, and assign shaders for internal stages (see BlurSharpenFilter for an example).
        #
        # If your filter has no internal stages, it does not need to define the make() function.

    def synthesize(self):
        # Generate fshader code. See docstring of this function for details.

    def update(self, task=None):
        # Update those run-time parameters that must be updated every frame.
        # If your filter does not have any such parameters, then there is no need
        # to provide this function.
        #
        # This is added to taskMgr by FilterStage at the end of reconfigure(),
        # if self.registerUpdatable() has been called in setup().

    # Create properties for filter parameters using Filter.makeRTParam() and Filter.makeCTParam()
    # property factories.

    # property "param1" maps to run-time parameter "mycooleffect_param1" for compositing fshader
    param1 = Filter.makeRTParam("_param1", "param 1 (run-time)", "finalQuad", "mycooleffect_param1")

    # property "param2" maps to run-time parameter "mycoolparam" for interQuads[0]
    param2 = Filter.makeRTParam("_param2", "param 2 (run-time)", "interQuads", "mycoolparam", targetIdx=0)

    # property "param3" is a compile-time parameter, affecting the FilterStage's compositing fshader
    # (use the param in your shader generation code in synthesize())
    param3 = Filter.makeCTParam("_param3", "param 3 (compile-time)", level="stage")

"""

# TODO other imports
from __future__ import print_function
import inspect

def _isproperty(obj):
    return isinstance(obj, property)

def _indent(s, n):
    """Helper function to indent each line of a string.

    Supports single-line (no terminating linefeed), single-line (with terminating linefeed)
    and multi-line strings. Linefeeds will be preserved.

    Parameters:
      s = string to indent
      n = number of spaces to indent by

    Return value:
      indented string

    """
    hadTermLF = (len(s) > 0) and (s[-1] == "\n")
    # This unfortunately strips one terminating linefeed if the string has any...
    temp = "\n".join((n * " ") + line for line in s.splitlines())
    # ...so we add it back.
    if hadTermLF:
        return temp + "\n"
    else:
        return temp


"""Logical stages of the simulated image-forming process; ordered list of names considered valid.

Filters are assigned to these. Depending on properties of enabled filters, several stages may be
merged into a single render pass (FilterStage) to minimize the number of render passes.

"""
STAGES = ["SceneOptics", "LensFocus", "LensOpticsEarly", "LensOpticsLate", "FilmOrDetector", "Postprocessing", "DisplayDevice", "DebugHelpers"]

# Examples of filters belonging to different stages:
#   Preprocess:      Antialias [if added later], CartoonInk
#   SceneOptics:     LocalReflection [if added later], AmbientOcclusion, VolumetricLighting
#   LensFocus:       BlurSharpen (depends on interpretation), DepthOfField [if added later]
#   LensOpticsEarly: LensDistortion
#   LensOpticsLate:  Bloom, LensFlare   (to use distorted color texture as source)
#   FilmOrDetector:  Desaturation, Inverted (depends on interpretation)
#   Postprocess:     Tint
#   DisplayDevice:   GammaCorrection, Scanlines
#   DebugHelpers:    ViewGlow


"""Scene textures supported by FilterManager."""
SCENETEXTURES = ["color", "depth", "aux"]


def makeRTParam(dataMemberName, docString, targetName, shaderInputName, targetIdx=None):
    """Property factory. Add a run-time parameter to a filter.

    This function provides a "property template" for run-time filter parameters.
    It creates a property that comes with the necessary additional logic to behave
    as a run-time filter parameter.

    Usage:

        class MyFilter(Filter):
            def __init__(self, **kwargs):
                # ...

            # ...methods...

            param1 = Filter.makeRTParam("_param1", "my parameter description", "finalQuad", "myfilter_param1")

    Note that the call to Filter.makeRTParam() belongs in the class scope, just like a property() declaration.

    The variable name, as it is accessed in the shader code, can be either "myfilter_param1"
    or (using the old convention) "k_myfilter_param1"; setShaderInput(), which the
    property setter will call internally, expects the name without the "k_" prefix (if any).

    What the above usage example does:

        - An property called "param1" is created for MyFilter instances.

        - "_param1" is used as the name of the underlying data member.

        - The data member is created at the time of first write to the property.
          It is invalid to read the property before first writing to it at least once
          (i.e. filter parameters follow the same semantics as regular data members).

        - When you assign a value to param1 (in the code using MyFilter), the setter automatically
          checks that the filter is attached (i.e. the quad having the shader exists) before it tries
          to send the shader input to FilterManager.

          If the filter is not attached, the data member is updated, but no shader input is sent.
          The attach() method of Filter takes care of nudging all setters so that up-to-date
          configuration is sent to the shaders at filter attach time (i.e. when the shaders
          become available).

        - Which quad the input is sent to, and under which shader input name, are controlled by
          the arguments targetName and shaderInputName.

          If you use the "k_" prefix in naming custom shader inputs, you MUST strip it from
          shaderInputName given to this function.

    The parameter targetName specifies either the name of a quad reference stored to self,
    or the name of a list of quads stored to self.

    Usually this is either the string "finalQuad" (meaning self.finalQuad), or the string
    "interQuads" (meaning self.interQuads). In the case of a list (interQuads), the parameter
    targetIdx selects a quad from the list.

    For example,

        param1 = Filter.makeRTParam("_param1", "my parameter description",
                                    "interQuads", "myfilter_param1", targetIdx=1)

    defines that the setter should send the value of param1 into the shader input named
    "myfilter_param1" in the shader assigned to self.interQuads[1].

    For creating the interQuads and assigning shaders to them, see Filter.setup().
    A complete example is given in BlurSharpenFilter.

    """
    def _get(self):
        return getattr(self, dataMemberName)

    def _set(self, value):
        # Generic run-time filter parameter setter.
        #
        # We always store the new data value, but only send shader input
        # if the quad (which has the shader) exists when we are being called.
        #
        setattr(self, dataMemberName, value)

        # Produce reference to the target that is to receive the shader input.
        #
        # We need to be rather paranoid here, because we are accessing attributes
        # that are updated at run time. Especially interQuads is typically
        # updated in user-defined code (in setup() of filters that require
        # intermediate stages).
        #
        if targetIdx is None:
            targetRef = getattr(self, targetName)
            if type(targetRef) is list:
                raise TypeError("target 'self.%s' is a list, but no targetIdx=... was given" % targetName)
            completeName = "self.%s" % targetName
        else:
            targetRefs = getattr(self, targetName)
            if type(targetRefs) is not list:
                raise TypeError("target 'self.%s' is a %s, not a list, but targetIdx=%d was given" % (targetName,
                                                                                                      type(targetRefs),
                                                                                                      targetIdx))
            targetRef = targetRefs[targetIdx]
            completeName = "self.%s[%d]" % (targetName, targetIdx)

        # If a NodePath is currently stored in targetRef, send the new value to the specified shader input.
        #
        if targetRef is not None:
            # At this point, ref should point to a NodePath that has our shader.
            if not isinstance(targetRef, NodePath):
                raise TypeError("target '%s' must be a NodePath instance, but is a %s" % (completeName, type(targetRef)))

            ref.setShaderInput(shaderInputName, value)

    return property(_get, _set, None, docString)


def makeCTParam(dataMemberName, docString, level="stage"):
    """Property factory. Add a compile-time parameter to a filter.

    This function provides a "property template" for compile-time filter parameters.
    It creates a property that comes with the necessary additional logic to behave
    as a compile-time filter parameter.

    Parameters:

      dataMemberName = string. Name for the internal data member that will be
                       created and used to store the value.

      docString      = string. Documentation for this compile-time parameter.

      level          = string. Determines which level of the filter system
                       this compile-time parameter affects.

                       One of:

                         "filter"          = Only the shaders of internal stages of this Filter are affected.

                         "stage" [default] = The compositing shader of the FilterStage is affected.

                         "pipeline"        = The whole FilterPipeline is affected.
                                             (E.g. stageName and isMergeable of the Filter base class.)

    The level information is used to direct the reconfigure() call to the appropriate level,
    when the value of the compile-time parameter changes.

    Usage:

        class MyFilter(Filter):
            def __init__(self, **kwargs):
                # ...

            # ...methods...

            param2 = Filter.makeCTParam("_param2", "my parameter description")

    Note that the call to Filter.makeCTParam() belongs in the class scope, just like a property() declaration.

    What the above usage example does:

        - An property called "param2" is created for MyFilter instances.

        - "_param2" is used as the name of the underlying data member.

        - The data member is created at the time of first write to the property.
          It is invalid to read the property before first writing to it at least once
          (i.e. filter parameters follow the same semantics as regular data members).

        - When you assign a value to param2 (in the code using MyFilter), the setter automatically
          flags a compile as needed if the value is different from the stored one (and always at first write).

        - If the appropriate object is available, and compiles have not been disabled by beginBatchUpdate(),
          the compile then runs immediately. If the compile runs, it clears the "needs compile" flag.

          In the case of a batch update, the batchUpdateDone() hook triggers the pending compile, if any.
          The flag status checking is done inside the appropriate reconfigure(), however, because
          batchUpdateDone() is not its only caller.

    """
    # Raise define-time parameter validation errors already during property creation.
    if level not in ["filter", "stage", "pipeline"]:
        raise ValueError("Compile-time parameter level '%s' unknown; valid: 'filter', 'stage' [default] and 'pipeline'." % level)

    def _get(self):
        return getattr(self, dataMemberName)

    def _set(self, value):
        # Generic compile-time filter parameter setter.
        #
        # We avoid unnecessary recompiles by only flagging a compile:
        #
        #  - during first write
        #  - when the value actually changes
        #
        # Note that if a Filter is not currently attached to a FilterStage or a
        # FilterPipeline, the _needsCompile flag at that level will not be set.
        #
        # However, "not attached" also means exactly what it says on the tin,
        # so this is correct. If the Filter is currently not part of a FilterStage
        # or a FilterPipeline, then obviously no FilterStage or FilterPipeline
        # needs to be recompiled, even if the filter's compile-time parameters
        # (having those levels) change.
        #
        # It is assumed that the objects eventually requesting the attach are intelligent enough
        # to set their own _needsCompile flag when they attach a filter, as the attach operation
        # itself changes their compile-time configuration (because of the new filter added;
        # the compositing shader needs to be recompiled, or possibly the whole pipeline rebuilt).
        #
        if (not hasattr(self, dataMemberName))  or  value != getattr(self, dataMemberName):
            if level == "filter":
                self._needsCompile = True
            elif level == "stage":
                if self.filterStage is not None:
                    self.filterStage._needsCompile = True
            else: # level == "pipeline":
                if self.pipeline is not None:
                    self.pipeline._needsCompile = True

        setattr(self, dataMemberName, value)

        # Trigger the recompile to apply changes immediately.
        #
        # (The recompile may be cancelled inside reconfigure() if our caller has called
        #  the corresponding beginConfigurationUpdate(). That mechanism is used when
        #  the caller knows that several properties will be batch-updated.)
        #
        if level == "filter":
            if self._needsCompile:
                self.reconfigure()
        elif level == "stage":
            if self.filterStage is not None  and  self.filterStage._needsCompile:
                self.filterStage.reconfigure()
        else: # level == "pipeline":
            if self.pipeline is not None  and  self.pipeline._needsCompile:
                self.pipeline.reconfigure()

    return property(_get, _set, None, docString)


def makeUTParam(dataMemberName, docString):
    """Property factory. Add an update-time parameter to a filter.

    This is used for creating parameters that control the internal logic of the user-defined update() method.

    See ScanlinesFilter for an example.

    """
    # Currently, update-time parameters do no additional processing beside setting the internal data member.
    # This just allows code that is shorter and more similar to other kinds of parameter definitions
    # than declaring the property manually.

    def _get(self):
        return getattr(self, dataMemberName)

    def _set(self, value):
        setattr(self, dataMemberName, value)

    return property(_get, _set, None, docString)


def makeFilterTexture(texName):
    """Helper function to create fullscreen filter textures, setting some appropriate parameters automatically."""
    tex = Texture("scene-"+texName)
    tex.setWrapU(Texture.WMClamp)
    tex.setWrapV(Texture.WMClamp)
    return tex


class BatchUpdatable(object):
    """Base class for objects that support batch updates.

    Starting a batch update typically temporarily disables some expensive operation related to
    configuration changes, which is then re-enabled (and applied) when the caller informs
    that the batch is complete.

    What a batch update means in practice for each derived class, and how to behave when
    a batch update is (or is not) in progress, is left for that derived class to define.
    This class simply provides the general support logic.

    In the case of Filter, FilterStage and FilterPipeline, a batch update allows updating several properties
    (which may contain compile-time parameters, which affect shader code generation) without triggering
    unnecessary intermediate recompiles after every property change.


    There are three sets of methods in BatchUpdatable:

      beginBatchUpdate() / endBatchUpdate():
        Interface methods, responsible for routing the update notification toward higher levels in the
        object hierarchy, when the user starts a batch update on some object instance located anywhere
        in the hierarchy.

        Overridable. The default is no hierarchy, just calling self.beginBatchUpdateImpl() or
        self.endBatchUpdateImpl(), respectively.

        If overridden, this should either call beginBatchUpdate() / endBatchUpdate() of an object
        higher in the hierarchy - or if the desired "root level" is accessible - beginBatchUpdateImpl()
        or endBatchUpdateImpl() (respectively) of that "root level" object.

        At the root level of the hierarchy, the default implementation does the right thing.

      beginBatchUpdateImpl() / endBatchUpdateImpl():
        Implementation methods, responsible for managing the local counter (for the object itself),
        entering/exiting batch update mode as needed, and calling hooks to propagate the ...Impl()
        call to any contained BatchUpdatables (objects at lower levels of the hierarchy).

        Do not override.

      beginBatchUpdateImplHook() / endBatchUpdateImplHook():
        Hooks for beginBatchUpdateImpl() / endBatchUpdateImpl(), meant for passing the call to any
        BatchUpdatables contained by self.

        Override these to provide the hook code. The hook should call beginBatchUpdateImpl() or
        endBatchUpdateImpl() (respectively) of the appropriate objects contained by self.

        At the leaf level of the hierarchy, the default implementation (which is blank)
        does the right thing.


    In an object hierarchy, the idea is to route the call through a "V cycle". Each level propagates
    the batch update notification toward higher levels, until the notification hits the root.
    Then each level (starting from the root) applies the batch update notification to all of its
    contained objects (objects lower in the hierarchy).

    This allows for the user (or an internal mechanism) to call beginBatchUpdate()/endBatchUpdate()
    for any object in the hierarchy, causing the whole hierarchy to enter/exit batch update mode.
    In the postprocessing filter system, this is especially useful, because compile-time parameters
    in the Filters may require any of Filter, FilterStage or FilterPipeline recompiles when their
    values are changed; especially updateConfigurationFromKeywords() wants to avoid unnecessary
    recompiles while more properties are still being set.

    """
    def __init__(self):
        """Constructor."""
        self._batchUpdateCounter    = 0
        self._batchUpdateInProgress = False


    @property
    def batchUpdateInProgress(self):
        """Whether a batch update is currently in progress. Read-only property.

        Derived classes are expected to query this property to decide whether to take action,
        when that action is something that should be disabled during a batch update.

        """
        return self._batchUpdateInProgress

    def beginBatchUpdate(self):
        """Begin a batch update.

        Interface method.

        Responsible for delegating the call to the next higher level in the object hierarchy.
        Default is to just run self.beginBatchUpdateImpl().

        Override this if you need trigger hierarchical updates from a lower-level object.

        """
        self.beginBatchUpdateImpl()

    def beginBatchUpdateImpl(self):
        """Begin a batch update.

        Implementation method.

        It is valid to call beginBatchUpdate() several times in a nested manner, if needed.
        In this case, the batch update is considered complete after equally many calls to
        endBatchUpdate(). This is implemented using a counter.

        """
        self._batchUpdateCounter += 1
        self._batchUpdateInProgress = True
        self.beginBatchUpdateImplHook()

    def beginBatchUpdateImplHook(self):
        """Hook for user-defined code supplied by derived classes. Called at end of beginBatchUpdateImpl().

        Useful for propagating a batch update to any other BatchUpdatables contained by self.

        """
        pass

    def endBatchUpdate(self):
        """End a batch update.

        Interface method.

        Responsible for delegating the call to the next higher level in the object hierarchy.
        Default is to just run self.endBatchUpdateImpl().

        Override this if you need trigger hierarchical updates from a lower-level object.

        """
        self.endBatchUpdateImpl()

    def endBatchUpdateImpl(self, notify=True):
        """End a batch update.

        Implementation method.

        Parameters:

            notify = bool. Set this to False to prevent calling batchUpdateDone().

                           (Usually no need, but can be useful under special circumstances.
                            This is used internally by Filter to prevent infinite recursion
                            during FilterStage attach() time, because its batchUpdateDone()
                            triggers a FilterPipeline.reconfigure() (due to propagation),
                            but attach() is called from FilterPipeline.reconfigure().)

        """
        self.endBatchUpdateImplHook(notify)
        self._batchUpdateCounter -= 1
        if self._batchUpdateCounter < 1:
            self._batchUpdateCounter = 0
            self._batchUpdateInProgress = False
            if notify:
                self.batchUpdateDone()

    def endBatchUpdateImplHook(self, notify):
        """Hook for user-defined code supplied by derived classes. Called at start of endBatchUpdateImpl().

        Useful for propagating a batch update to any other BatchUpdatables contained by self.

        The notify flag is passed through from endBatchUpdate().

        """
        pass

    def batchUpdateDone(self):
        """Hook for notifying a derived class when a batch update is complete.

        Override this to implement your deferred processing when a batch completes.

        Note:

          - endBatchUpdateImplHook() is always called, at start of endBatchUpdateImpl(),
            regardless of the value of the counter.

          - batchUpdateDone() is called only when the counter has reached zero,
            at *end* of endbatchUpdateImpl(), and only when notify=True
            (in the call to endBatchUpdateImpl()).
            This is the desired behavior for deferred processing.

        """
        pass


class Filter(BatchUpdatable):
    """Abstract base class for postprocessing filters. Filters are implemented in derived classes."""

    def __init__(self, **kwargs):
        """Constructor.

        This calls self.detach() and self.resetConfiguration(), and loads any parameters from kwargs
        into properties.

        The key of each kwarg must be the name of a valid property of the instance being initialized.
        Trying to set a property that does not exist in the current instance (current filter type)
        raises AttributeError; this is meant to protect against human error in the parameter list
        (typos, and trying to send settings that belong to a different type of filter).

        This abstract base class defines the pipeline-related properties: "sort", "stageName" and "isMergeable".
        These must usually be set in reset() of derived classes; "sort" does not even have a
        valid default value, as the sort value is very filter-specific.

        In your derived classes, be sure to call super(self.__class__, self).__init__().

        Exact placement of the call (within derived class's __init__()) does not matter,
        as long as you keep in mind that this will basically run detach() (to create the
        related data members and initialize them to empty values), resetConfiguration(),
        and updateConfigurationFromKeywords().

        """
        # batch update mechanism
        super(self.__class__, self).__init__()

        # The _needsCompile flag stores whether this filter instance currently needs to be (re)compiled.
        #
        # It tracks changes to Filter-level compile-time parameters; i.e. those affecting
        # the shaders of internal stages.
        #
        self._needsCompile = True

        # self.pipeline and self.filterStage are managed by attach()/detach().
        #
        # We set these to None before calling detach() to create the data members and
        # initialize them with blanks, because cleanup() in derived classes (called from detach())
        # might reference these.
        #
        # (It makes possible errors in user-defined code easier to debug if the attribute always exists;
        #  something being None is a clear indication of that something not being initialized, while a
        #  missing attribute looks like an internal bug.)
        #
        self.pipeline    = None
        self.filterStage = None
        self.task        = None
        self.cleanup()

        # This will call reset().
        #
        self.resetConfiguration()

        # Because we are not yet attached to any FilterStage, the setters should do the right thing
        # and just set the corresponding internal variable to the new value.
        #
        self.updateConfigurationFromKeywords(**kwargs)

    def __del__(self):
        """Destructor."""
        self.cleanup()

    def cleanup(self):
        """Cleanup method for whole-hog object cleanup.

        (For the opposite of setup(), see setdown().)

        """
        self.detach(phase="stage")
        self.detach(phase="pipeline")


    def ls(self, indent=0):
        """Print a human-readable description of this Filter instance into the terminal."""
        ind = (indent * " ")
        nints = len(self.interQuads)
        if nints == 0:
            internal_passes = ""
        else:
            intplural = "es" if nints != 1 else ""
            internal_passes = "; %s internal render pass%s" % (nints, intplural)
        print("%s%s instance at 0x%x%s" % (ind, self.__class__.__name__, id(self), internal_passes))
        if len(self.textures) > 0:
            # sort by texture name for human-readability
            print("%s  Internal textures: %s" % (ind, sorted(self.textures.keys())) )
        print("%s  Configuration:" % ind)
        propnames = self.getProperties()
        conf      = self.getConfigurationAsKeywords()
        for p in propnames:
            print("%s    %s: %s" % (ind, p, conf[p]))


    # It is convenient to implement the sort value as a compile-time parameter, because changing
    # its value implies that the compositing fshader must be re-synthesized (i.e. the filter stage
    # needs a recompile).
    #
    sort = Filter.makeCTParam("_sort",
        """Sort value of this filter inside the pipeline stage it resides in.

        Integer, 0 <= sort < 100. Smaller means earlier. A special value of -1 is reserved
        for the pipeline stage initialization filter (see StageInitializationFilter).

        The sort value controls the ordering of the filters within the compositing shader
        of the pipeline stage.

        See also:

          stageName
          isMergeable

        """)

    # Changing stageName or isMergeable requires the whole pipeline to be recompiled,
    # as (at least) two different stages are affected.
    #
    # (May affect more than two stages, depending on exactly which enabled filters have isMergeable set.
    #  Hence the simplest correct solution is to fully rebuild the pipeline.)
    #
    stageName = Filter.makeCTParam("_stageName",
        """Which pipeline stage this filter should go into.

        String, one of Filter.STAGES. (Feel free to expand the list if needed when implementing new filters.)

        Each stage tentatively marks a new render pass (but see isMergeable).

        See also:

          sort
          isMergeable

        """, level="pipeline")

    isMergeable = Filter.makeCTParam("_isMergeable",
        """Whether this filter can be merged into a previous stage.

        Bool.

        If True, stageName is considered tentative. If the stage specified by stageName contains
        only mergeable filters at pipeline reconfigure() time, the stage will be eliminated 
        (to reduce the number of render passes), by placing all filters from it to a previous stage.

        If False, stageName is considered absolute. This filter will be placed in the
        named stage, and nowhere else. A new render pass will be created, if not already
        created for another filter with the same stageName.

        Default is True. This should be set to False only in those Filter subclasses that need
        access to up-to-date texture information outside the current pixel in the fshader.
        BlurSharpenFilter is an example.

        Note that the up-to-date color for the current pixel is always available in pixcolor.

        If your filter algorithm is such that it cannot respect previous changes to pixcolor,
        then it should begin a new stage, with stageName=..., isMergeable=False and sort=0.

        See also:

          sort
          stageName

        """, level="pipeline")

    def batchUpdateDone(self):
        """When a batch update completes, recompile shaders defined for internal stages."""
        self.reconfigure()

    def beginBatchUpdate(self):
        """Begin a batch update, accounting for the filter system hierarchy.

        If self.pipeline is available, begin batch update there.

        Otherwise, if self.filterStage is available, begin batch update there.

        Finally, if neither of those is available, begin batch update locally.

        """
        if self.pipeline is not None:
            self.pipeline.beginBatchUpdateImpl()  # this propagates to FilterStage and Filter
        elif self.filterStage is not None:
            self.filterStage.beginBatchUpdateImpl()  # this propagates to Filter
        else:
            self.beginBatchUpdateImpl()

    def endBatchUpdate(self, notify=True):
        """End a batch update, accounting for the filter system hierarchy.

        If self.pipeline is available, end batch update there.

        Otherwise, if self.filterStage is available, end batch update there.

        Finally, if neither of those is available, end batch update locally.

        """
        if self.pipeline is not None:
            self.pipeline.endBatchUpdateImpl(notify)  # this propagates to FilterStage and Filter
        elif self.filterStage is not None:
            self.filterStage.endBatchUpdateImpl(notify)  # this propagates to Filter
        else:
            self.endBatchUpdateImpl(notify)

    def resetConfiguration(self):
        """Set filter-specific settings to their default values.

        Interface method.

        Do not override this; instead, override reset(), which is provided for that purpose.

        This is an atomic operation; the pipeline is guaranteed not to recompile
        while this is in progress.

        """
        # Disable compiles.
        self.beginBatchUpdate()

        try:
            # Run the actual configuration reset code.
            #
            # This is essentially user-defined code, so we must use "finally" to ensure
            # that the batch update is ended even if an exception is raised.
            #
            self.reset()
        except:
            raise
        finally:
            # Re-enable compiles, triggering a compile immediately if a compile-time parameter has changed.
            #
            # At __init__() time, though, we are not yet attached to a FilterPipeline or FilterStage,
            # so at that point this will do nothing. (Note that even compiling the internal intermediate stage
            # shaders requires us to be attached to a FilterStage, because the setup of the interQuads requires
            # scene textures, which are not available until the Filter is attached to a FilterStage.)
            #
            # The compile disable/enable logic is here to account for resetConfiguration() calls that happen
            # later, for a running filter instance.
            #
            self.endBatchUpdate()

    def reset(self):
        """Set filter-specific settings to their default values.

        Implementation method.

        While this function is running, recompiles are guaranteed to be disabled.

        Derived classes should override this to set their sort, stageName and isMergeable values,
        and to set filter-specific parameters to sensible defaults.

        Be sure to call super(self.__class__, self).reset() *first* in your
        reset() to make sure that any inherited properties are reset correctly.
        This requirement holds whether your class inherits from Filter or from an existing derived class.

        Derived classes ARE allowed, if it is considered necessary, to set different default values
        for inherited properties than their ancestors.

        The Filter base class defines three properties: "sort", "stageName" and "isMergeable".

        Always be sure to reset the configuration by assigning to the PROPERTIES, not directly to their
        internal data members. This is important for Filter.makeRTParam() and Filter.makeCTParam()
        to do their automatic bookkeeping correctly.

        This also allows resetConfiguration() to behave as expected, not only during initialization,
        but also when called for a running filter instance.

        """
        # Note we set properties, not their internal data members.
        self.sort        = None  # no valid default
        self.stageName   = None  # no valid default
        self.isMergeable = True

    def updateConfigurationFromKeywords(self, **kwargs):
        """Set zero or more filter-specific settings, given as kwargs.

        Any setting that is implemented as a property (especially Filter.makeRTParam() and
        Filter.makeCTParam() properties) is supported.

        Trying to set a property that does not exist in the current instance (current filter type)
        raises AttributeError; this is meant to protect against human error in the parameter list
        (typos, and trying to send settings that belong to a different type of filter).

        Any properties not specified in kwargs retain their previous values.

        There is no need to override this in derived classes.

        This is an atomic operation; the pipeline is guaranteed not to recompile
        until all the specified properties have been updated.

        If the Filter is not currently attached to a FilterStage, this simply updates the
        stored data values; they will then be automatically applied at reconfigure() time,
        when attach() is called.

        See also:

          getConfigurationAsKeywords()

        """
        self.beginBatchUpdate()

        try:
            propertynames = self.getProperties()
            for key,value in kwargs.items():
                if key not in propertynames:
                    raise AttributeError( "%s has no property named '%s'" % (self.__class__.__name__, key) )
                setattr(self,key,value)
        except:
            raise
        finally:
            self.endBatchUpdate()

    def getConfigurationAsKeywords(self):
        """Get all filter-specific settings.

        The return value is a dict, mapping property names (as strings) to the values
        they had at the time getConfigurationAsKeywords() was called.

        This can be used to store and later reload filter configuration values
        (assuming that no mutable objects are stored in the properties).

        If you want to get the default values, you can call resetConfiguration()
        and then this.

        Example:

          # fpp is a FilterPipeline instance.
          #
          # Here we assume a MyCoolEffectFilter has been added to fpp earlier.
          #
          f = fpp.findFilter(MyCoolEffectFilter)

          # Store current configuration
          storedSettings = f.getConfigurationAsKeywords()

          # ...do stuff...

          # Reload stored configuration
          f.updateConfigurationFromKeywords(**storedSettings)

          # Add a copy of the filter with the same configuration into
          # another pipeline instance (such as when rendering
          # two independent views)
          #
          myotherfpp.addFilter(MyCoolEffectFilter, **storedSettings)

        See also:

          updateConfigurationFromKeywords()

        """
        result = {}
        propertynames = self.getProperties()
        for key in propertynames:
            result[key] = getattr(self, key)
        return result

    def loadShader(self, filename):
        """Convenience method to load a Cg shader .sha file placed in the same directory as Filter.py.

        This is OS-agnostic and TrueCases the filename automatically.

        Parameters:

          filename = string. Filename without the path.

        Return value:

          Passed through from Shader.load().

        """
        fn = os.path.join(os.path.abspath(os.path.dirname(__file__)), filename)
        fn = Filename.fromOsSpecific(fn)
        fn.makeTrueCase()

        shaderObj = Shader.load(fn)
        if shaderObj is None:
            self.cleanup()
            raise RuntimeError("Shader.load() for shader file '%s' failed in %s" % (filename, self.__class__.__name__))
        return shaderObj

    def getProperties(self):
        """Return property names defined in the current instance.

        A sorted list is returned.

        See also getPropertiesWithDocs().

        """
        propertynames = [name for (name, value) in inspect.getmembers(self.__class__, _isproperty)]
#        # For future extension:
#        propertynames = filter( lambda s: s[0] != "_", propertynames )  # ignore internal properties
        propertynames.sort()
        return propertynames

    def getPropertiesWithDocs(self):
        """Return a list of property names defined in the current instance, along with their docstrings.
    
        This can be used to quickly extract a human-readable description of the parameters supported
        by a given filter.

        Return format is a sorted list of tuples (name, docstring).

        """
        properties = inspect.getmembers(self.__class__, _isproperty)
        result = []
        for name,prop in properties:
#            # For future extension:
#            if name[0] == "_":
#                continue  # ignore internal properties
            result.append( (name, getattr(prop,"__doc__")) )
        result.sort(key=lambda obj: obj[0])
        return result

    def attach(self, pipeline=None, filterStage=None):
        """Internal function. Attach the filter to a FilterPipeline or a FilterStage.

        This is called internally from FilterPipeline.reconfigure()
        and FilterStage.reconfigure().

        Exactly one of pipeline or filterStage must be not None.
        Correct call sequence is to attach pipeline first, then filterStage.

        There is no need to override this in derived classes; instead, see setup().


        During a pipeline attach, this calls setup() for the "pipeline" phase
        (for registering input textures, custom shader inputs and aux bits),
        and stores the pipeline reference.

        Having a stored pipeline reference means that updates to pipeline-level
        compile-time parameters in this Filter will affect the pipeline;
        also any internal batch update start/end calls will be directed
        to the pipeline.


        During a filterStage attach, this configures the FilterStage input,
        calls setup() for the "stage" phase, and nudges all property setters
        to run, updating shader inputs to the current run-time parameter values.


        This function is separate from the constructor, because filter objects
        are persistent across pipeline reconfigures.


        Parameters:

            pipeline    = FilterPipeline instance where this Filter instance is inserted
                          (i.e. the FilterPipeline instance whose scene texture and
                           and aux bits requirements this filter affects).

            filterStage = FilterStage instance where this Filter instance is inserted
                          (i.e. the FilterStage instance whose compositing fshader
                           this filter goes into).

        See also:

          detach()
          setup()
          setdown()

        """
        # check parameter combinations
        if (pipeline is None  and  filterStage is None)  or  (pipeline is not None  and  filterStage is not None):
            raise ValueError("Exactly one of pipeline or filterStage must be not None; got pipeline=%s, filterStage=%s." % (pipeline, filterStage))

        # check parameter data types
        if pipeline is not None  and  not isinstance(pipeline, FilterPipeline):
            raise TypeError("pipeline must be an instance of FilterPipeline; got '%s'" % type(pipeline))
        if filterStage is not None  and  not isinstance(filterStage, FilterStage):
            raise TypeError("filterStage must be an instance of FilterStage; got '%s'" % type(FilterStage))

        # check call sequence constraints
        if filterStage is not None  and  self.pipeline is None:
            raise ValueError("Cannot attach to a FilterStage while not attached to a FilterPipeline. Attempted to attach to FilterStage '%s'." % filterStage.stageName)
        if pipeline is not None  and  self.filterStage is not None:
            raise ValueError("Cannot attach to a FilterPipeline while already attached to a FilterStage. Currently attached to FilterStage '%s'." % self.filterStage.stageName)
        if pipeline is not None  and  self.pipeline is not None:
            raise ValueError("Cannot attach to a FilterPipeline while already attached to a FilterPipeline.")
        if filterStage is not None  and  self.filterStage is not None:
            raise ValueError("Cannot attach to FilterStage '%s' while already attached to FilterStage '%s'." % (filterStage.stageName, self.filterStage.stageName))


        # do actual work
        if pipeline is not None:
            assert( self.filterStage is None )

            # Attaching to a FilterPipeline.
            #
            # This implies that batch update control will be connected to the pipeline,
            # and setting any pipeline-level compile-time parameters while attached
            # will recompile the pipeline.
            #
            self.pipeline = pipeline

            # Register input textures, custom shader inputs and aux bits.
            #
            self.setup(phase="pipeline")

        else: # filterStage is not None:
            # Attaching to a FilterStage.
            assert( self.pipeline is not None )
            assert( isinstance(self.pipeline, FilterPipeline) )

            self.filterStage = filterStage

            # finalQuad = NodePath instance; output quad of filterStage.
            #
            #             This is used to pass custom parameters to the compositing fshader
            #             via setShaderInput() (in the property setters in derived classes).
            #
            #             Note that it is the finalQuad of *this pipeline stage*.
            #             It might not be "the" finalQuad of the whole pipeline,
            #             unless this filter resides in the last stage of the pipeline.
            #
            # inputTextures  = Input texture dictionary, containing input textures
            #                  of filterStage. This is used to access the
            #                  pipeline stage input in filters containing internal stages.
            #
            #                  May contain entries keyed by names in Filter.SCENETEXTURES,
            #                  their values containing the references to the corresponding
            #                  Texture instances.
            #
            self.finalQuad      = filterStage.finalQuad
            self.inputTextures  = filterStage.stageInputTextures

            # Create any interQuads that are required by this filter.
            #
            # We do this when attaching to a FilterStage, because setting up the internal pipelining
            # for the internal stages requires the pipeline stage input textures to be available.
            #
            # Important to do this first before applying configuration; the interQuads will contain shaders,
            # which may be the target for some of the shader inputs in run-time parameters.
            #
            self.setup(phase="stage")

            # Compile internal shaders and assign them to interQuads.
            #
            # (This is done in user code, in make().)
            #
            self._needsCompile = True
            self.reconfigure()

            # Run all property setters to apply the current configuration.
            #
            # We need to do this because only each property setter itself knows
            # what needs to be done when the value of the property changes.
            #
            # Generally, they either set shader inputs (either to the compositing shader
            # of the pipeline stage, in self.finalQuad, or to an internal intermediate shader
            # in self.interQuads[idx] for some value of idx), or trigger a recompile
            # (if the value of a compile-time parameter has changed).
            #
            # What we actually want to do here is to update the shader inputs.
            #
            self.beginBatchUpdate()  # at this point we always have self.pipeline

            for s in self.getProperties():
                # Loopback to keep current value, while forcing setter side effects.
                setattr(self,s, getattr(self,s))

            # We are called only from FilterStage.reconfigure(); thus a compile is
            # already in progress. Prevent infinite recursion by *not* triggering
            # a new compile.
            #
            # (Nothing should happen, anyway, since we just looped back the values;
            #  the compile only triggers if the value of a compile-time parameter
            #  actually changes.)
            #
            self.endBatchUpdate(notify=False)

    def setup(self, phase):
        """Filter-specific setup hook. Called from attach().

        Abstract method. Must be overridden in derived classes.

        Parameters (provided by attach()):

          phase = string. One of:

                    "pipeline" = This Filter is now being attached to a FilterPipeline.

                    "stage"    = This Filter is now being attached to a FilterStage.

        The implementation must:

          Always:

            - If the class inherits from an existing filter (not from the abstract Filter base class),
              call super(self.__class__, self).setup(phase).

              If the class inherits directly from Filter, then it MUST NOT call super(...).setup(phase),
              since this method is abstract in the base class, and calling it will raise NotImplementedError.

          In phase="pipeline":

            - Register input textures and custom inputs needed by this filter in the compositing fshader.
              Use self.registerInputTexture() and self.registerCustomInput() to do this.

            - Register aux bits, if needed. Use self.registerAuxBits().

          In phase="stage":

            - Allocate resources for internal processing stages, if needed.

              For example, a blur filter might have internal "blur-x" and "blur-y" passes
              that are applied to the pipeline stage input, and then their output is used
              in the compositing fshader of the pipeline stage.

              See also Filter.make().

              See BlurSharpenFilter for a complete example.

        See also:

          attach()
          detach()
          setdown()

        """
        raise NotImplementedError("Abstract method setup() must be implemented in classes derived from Filter")

    def detach(self, phase):
        """Detach the filter from the FilterStage or the FilterPipeline.

        Parameters:

          phase = string. One of:

                    "stage"    = Detach the filter from the FilterStage.

                    "pipeline" = Detach the filter from the FilterPipeline.

        During a stage detach, this calls self.setdown() for the "stage" phase to run any custom cleanup,
        sets the pipeline stage input as invalid, and cleans up any internal processing stages.
        The internal stages are cleaned up by setting self.textures and self.interQuads to blank.

        During a pipeline detach, this calls self.setdown() for the "pipeline" phase,
        and discards the pipeline reference. (This implies that batch update control will be disconnected
        from the pipeline, and setting any pipeline-level compile-time parameters while not attached
        will not recompile the pipeline.)

        See also:

          attach()
          setup()
          setdown()

        """
        # This function must not raise exceptions; it is called during cleanup,
        # which may be called in response to exceptions.
        assert( phase in ["pipeline", "stage"] )

        if phase == "stage":
            if self.task is not None:
                taskMgr.remove(self.task)
                self.task = None
            self.setdown(phase="stage")
            self.textures      = {}
            self.interQuads    = []
            self.finalQuad     = None
            self.inputTextures = {}
            self.filterStage   = None
            self._needsCompile = False  # not attached to anything; no need to compile
        else: # what == "pipeline":
            self.setdown(phase="pipeline")
            self.pipeline = None

            # This is used for logical stage -> render pass mapping support in FilterPipeline.
            #
            # This counts by how many stages the filter has been bumped backward during the pipeline build,
            # if the filter is mergeable (see FilterPipeline._createFilterStages()).
            #
            # This and self.sort will be combined to a final sort value when the filters are sorted
            # within FilterStage (see FilterStage.reconfigure()).
            #
            self.mergings = 0

    # The name "cleanup" is reserved for a whole-hog instance cleanup method,
    # to conform with the other objects.
    #
    def setdown(self, phase):
        """Filter-specific cleanup hook. Opposite of setup(). Called from detach().

        This can be overridden in derived classes to perform any filter-specific cleanup.
        Internal stages, if any, are automatically cleaned up by detach().

        When detach() is called at __init__() time to create the data members and initialize
        them with blanks, self.pipeline and self.filterStage will be None.

        Parameters (provided by detach()):

          phase = string. One of:

                    "stage"    = This Filter is now being detached from the FilterStage.

                    "pipeline" = This Filter is now being detached from the FilterPipeline.

        If setdown() is overridden, the implementation:

          - Must free any dynamically allocated resources created in its overridden setup()
            for the corresponding phase.

            Except: the most common resources needed by internal stages are cleaned up automatically.
            detach() for the "stage" phase already sets self.textures and self.interQuads to blank,
            and FilterPipeline already calls manager.cleanup() to delete the actual resources.

            If that is all you need, then your class does not need to provide a setdown() function.
            This is the most common case.

          - Must not raise exceptions; this is a cleanup function. Cleanup may be called in response
            to exceptions.

          - Must not touch configuration parameters (properties)! This allows Filter instances to
            retain their configuration across pipeline reconfigures.

          - If the class inherits from an existing filter (not from the abstract Filter base class),
            it must call super(self.__class__, self).setdown(phase) to ensure cleanup of any inherited
            dynamic resources.

        See also:

          attach()
          detach()
          setup()

        """
        pass


    def registerInputTexture(self, texName, needTexpix=False):
        """Register an input texture for the compositing fshader.

        Registered input textures will be inserted to the parameter list of the compositing fshader
        of the pipeline stage. They will also be inserted to the Cg function call for the filter
        that requested the register.

        Call this from setup() in derived classes to request textures to be provided to the compositing fshader.


        Note that since texture sizes may be different (e.g. internal stages can use
        half-resolution textures if they want to), the needTexpix flag is texture-specific.

        Including "color", "depth" or "aux" in the list in at least one filter that is enabled
        in the pipeline instructs the pipeline to configure itself to generate them.

        If at least one input texture is enabled in at least one filter in a FilterStage,
        l_texcoord will be available in the compositing fshader.


        Parameters:

          texName     = string. Name of texture to register. May be one of "color", "depth" or "aux"
                        (which are always available from the pipeline), or any internal texture
                        defined by the filter itself.

                        The latter is useful for filters with internal stages, which need (one or more of)
                        their intermediate textures as input to the compositing fshader.

                        In the case of an internal texture, the name must match the key used in
                        self.textures (in the Filter subclass object), i.e. the texture name
                        given to Filter.createInternalTextures().

                        If you intend to use the "aux" (fragment normals) texture, be sure to set also
                        the ABOAuxNormal aux bit (see registerAuxBits()).

          needTexpix  = bool. Whether texpix (pixel size) is needed for the texture
                        being registered. This is a per-texture setting, because each
                        needed texpix means sending one more input to the compositing fshader.

                        It also affects the generation of the parameter list of the Cg function
                        containing the fshader code for this filter.


        Examples:

          Many filters want the color input texture, but are not interested in the pixel size:

            self.registerInputTexture("color")

          Some filters want both color and depth:

            self.registerInputTexture("color")
            self.registerInputTexture("depth")

          A scanlines filter, which needs the pixel size in order to compute the scanline thickness
          in texture coordinate units, might want to do:

            self.registerInputTexture("color", need_texpix=True)

          A cartoon inking filter might be interested in the fragment normals, with pixel size
          to compute whole-pixel offsets for normal map analysis:

            self.registerInputTexture("aux", need_texpix=True)
            self.registerAuxBits( AuxBitplaneAttrib.ABOAuxNormal )

        """
        # The user-defined setup(), which calls us, is triggered by FilterPipeline, so we can use asserts.
        assert( self.pipeline is not None )
        assert( isinstance(self.pipeline, FilterPipeline) )
        self.pipeline._registerInputTexture(filterInstance=self, texName, needTexpix)

    def registerCustomInput(self, inputType, inputName):
        """Register a custom shader input for the compositing fshader.

        Registered custom inputs will be inserted to the parameter list of the compositing fshader
        of the pipeline stage. They will also be inserted to the Cg function call for the filter
        that requested the register.

        Call this from setup() in derived classes to request your custom shader inputs to be created
        for the compositing fshader.


        Custom inputs for any internal stages must NOT be registered; this function is to be used
        only for inputs to the compositing fshader.

        When naming your custom inputs, keep in mind that potentially a large number of different filters
        may get synthesized into the same pipeline stage. The parameter names within each pipeline stage
        MUST be unique. One way to ensure this is to use the naming convention k_filtername_parametername.


        Note that custom input handling is performed in two separate parts:

          - Registration ensures that code for receiving the input is composed and inserted to the
            generated shader.

          - A run-time property supplies the data to the input (via the property's setter).
            This occurs at attach() time, and while the filter is running, at any time
            when the run-time property is written to.


        Parameters:

          inputType = string. Low-level (Cg) datatype of the input. int, float, float4, ...

          inputName = string. The name of the parameter, exactly as it appears in the
                      compositing fshader code. (Do NOT strip the "k_" prefix if you use it.)

        Example:

          A cartoon ink filter might want to do:

            self.registerCustomInput("float4", "k_cartooncolor")

        """
        # The user-defined setup(), which calls us, is triggered by FilterPipeline, so we can use asserts.
        assert( self.pipeline is not None )
        assert( isinstance(self.pipeline, FilterPipeline) )
        self.pipeline._registerCustomInput(filterInstance=self, inputType, inputName)

    def registerAuxBits(self, bitmask):
        """Register aux bits.

        Call this from setup() in derived classes to request aux data such as normals or the glow map.

        
        Aux bits indicate what additional information the filter wants from the main renderer,
        in conjunction with the pipeline stage input textures. How the information is passed
        depends on its type. Glow maps (ABOGlow) are passed in the alpha channel of the
        pipeline stage input color texture, while normal maps (ABOAuxNormal) are passed as an
        additional texture (called "aux").

        ABOAuxNormal must be enabled for the main renderer to generate the "aux" texture
        containing fragment normals. (Additionally, you'll want to register "aux" as an input texture.)

        ABOGlow must be enabled to fill the alpha channel of the "color" texture with the glow map.


        The default (if this is not called) is 0, which means all aux bits are off.

        If you need to enable several aux bits in the same filter, bitwise-OR them together with "|".
        Aux bits requested by different filter instances are automatically ORed together by FilterPipeline.

        Parameters:

          bitmask = bitmask made of pandac.PandaModules.AuxBitplaneAttrib flags.

        """
        # The user-defined setup(), which calls us, is triggered by FilterPipeline, so we can use asserts.
        assert( self.pipeline is not None )
        assert( isinstance(self.pipeline, FilterPipeline) )
        self.pipeline._registerAuxBits(filterInstance=self, bitmask)

    def registerUpdatable(self):
        """Register this Filter as having shader inputs that must be updated automatically each frame.

        If you need this feature, call this from setup() in derived classes, and provide the
        actual update code by overriding update().

        """
        # The user-defined setup(), which calls us, is triggered by FilterPipeline, so we can use asserts.
        assert( self.pipeline is not None )
        assert( isinstance(self.pipeline, FilterPipeline) )
        self.pipeline._registerUpdatable(filterInstance=self, taskFunction=self.update)

    def synthesizeFragmentShader(self):
        """Synthesize and return fragment shader implementation.

        Interface method. There is no need to override this in derived classes.

        The return value is passed through from synthesize().

        """
        # Currently we don't need to do anything special here.
        # By having this method, we just reserve the possibility to add template code later.
        #
        return self.synthesize()

    def synthesize(self):
        """Synthesize and return fragment shader implementation.

        Abstract method.

        Override this in derived classes to provide the implementation of shader code generation.


        Return value:
          Tuple:
            (funcname, code, comment)
          where
            funcname = string, bare function name.

                       This function will be called in the compositing fshader of the FilterStage
                       where the Filter is attached.

            code     = string, implementation of the function, in Cg code.

                       Treat the top level as having zero indentation; top-level indentation
                       will be inserted by FilterStage when it writes the shader.

                       Terminate by a newline to make the resulting source more readable.

            comment  = string or None. Can be used to provide a short comment about what the function does.
                       This will be pasted at the call site (just before the function call) and
                       at the definition site (just before the definition of the function).
                       Must be a valid comment, i.e. either wrap it in /* */ or begin each line by // .

                       Must be terminated by a newline.

                       This should be one or two lines. For longer comments, it is recommended to
                       include them with the code, so that they will end up inside the function implementation.

                       Comments can be useful for debugging and examining the framework. Observe that
                       the source code of the currently active compositing shader can be retrieved
                       by reading the property FilterStage.shaderSourceCode.


        The pipeline automatically composes the final compositing fshader, using the shader code
        defined by synthesize() in the enabled filters. The function calls are ordered using the
        sort value of the filters.

        For each filter, the function declaration and the call to the function in the compositing fshader
        are generated automatically by the pipeline, using the data sent by registerInputTexture()
        and registerCustomInput() at setup() time.

        Explicitly:

          // ...comment of myFunction...
          float4 myFunction( /* params */,
                             float4 pixcolor )
          {
              // ...implementation of myFunction...

              return pixcolor;
          }

          void fshader( /* all params of current FilterStage */ )
          {
              // ...other code...

              // ...comment of myFunction...
              pixcolor = myFunction( /* params */,
                                     pixcolor );
        
              // ...other code...
          }

        The string "myFunction", the implementation of myFunction, and the comment of myFunction
        (or None if no comment) are what must be returned by synthesize(); all the rest is
        automatically generated, including the params list.

        Observe that fshader() expects that myFunction() will modify existing pixcolor,
        not completely overwrite it.

        See existing filters for examples.


        Conventions:

          The implementation must apply the current values of any relevant (FilterStage-level or
          FilterPipeline-level) compile-time parameters stored in the properties of self.

          All filters must treat the input color alpha channel as if it contains a glow map.
          This is to ensure correct operation also when AuxBitplaneAttrib.ABOGlow is enabled.

          The code must respect previous modifications to pixcolor in the same pipeline stage,
          to avoid overwriting results from other filters earlier in the stage. A filter for which
          this is not possible should be placed as the first one in the pipeline stage,
          by setting sort=0 and isMergeable=False.

          The implementation must explicitly call

            pixcolor = saturate(pixcolor);

          at the end if it wants to do that. The color is processed through a temporary variable
          to prevent unintended saturation while more processing is still being done. Only the final result
          (from the last filter in the FilterStage) is saturated automatically (this happens implicitly,
          by storing the result in the variable corresponding to the output color semantic in the fshader).

          Currently it is not possible to pass a non-saturated color from a FilterStage to the next FilterStage.

        """
        raise NotImplementedError("Abstract method synthesize() must be implemented in classes derived from Filter")

    def reconfigure(self, force=False):
        """Compile shaders for internal stages and assign them to interQuads.

        Interface method.

        This provides the logic to disable recompiles if currently in batch update mode,
        or if not currently attached to a FilterStage (in which case the interQuads do not exist),
        or if self._needsCompile is not set.

        This also clears the self._needsCompile flag if self.make() runs successfully.

        To implement the actual shader compilation code, override make().

        The interQuads must have been created in setup().


        Parameters:

          force = bool. If True, run the compile even if the self._needsCompile flag
                  is not set. This does not affect the other checks.

        See also:

          setup()
        
        """
        # Compiling the shaders for the interQuads requires that the Filter is attached
        # to a FilterStage, because only then the interQuads (to which the user code
        # will try to set the shaders) will exist.
        #
        # Trying to call reconfigure() when this requirement is not satisfied
        # is not an error, but the call will do nothing.
        #
        # Note that in such a case, self._needsCompile is left in its current state,
        # and hence a later reconfigure() (after self.filterStage becomes available)
        # may perform a compile that has been left pending earlier.
        #
        if self.filterStage is None:
            return

        # While batchUpdateDone() does indeed call reconfigure() (so the check may look silly),
        # also the setters of all individual Filter-level compile-time parameters do,
        # so we must indeed check the batch update flag before running the actual
        # reconfigure code.
        #
        if self.batchUpdateInProgress:
            return

        # Unless forced, compile only if a compile-time parameter has changed.
        if not force  and  not self._needsCompile:
            return

        self.make()
        self._needsCompile = False

    def make(self):
        """Compile shaders for internal stages and assign them to interQuads.

        Implementation method.

        Default implementation is blank; this is fine for filters which do not need internal stages.

        Override this method in derived classes if (and only if) you need internal stages
        in your filter.

        This method should load or (re-)compile the internal shaders, and assign them to the
        interQuads that have been created in setup(). The implementation must take into account
        the current values of the Filter-level compile-time parameters defined by that
        specific Filter (and any of its ancestors, if applicable; the abstract Filter base class
        itself has no such parameters).

        There is no need to check the state of, or set, any compile-related flags in your code;
        the interface method reconfigure() already does that, and only calls make()
        when safe and necessary.


        By design, this method has no return value.

        If an error occurs (e.g. renderQuadInto() fails), the implementation should raise
        an appropriate exception with a descriptive error message.

        On an error that can only arise if there is a strictly internal bug in the subclass's
        own code, the implementation should fail an assert.


        See also:

          setup()

        """
        pass

    def createInternalTextures(self, *args):
        """Create textures for internal stages.

        This is a convenience method, meant to be called from an overridden setup()
        in filters with internal stages.

        The created Texture objects will become available in self.textures,
        keyed by the names given.

        These same names can then be used, if necessary, in a registerInputTexture()
        call to configure the corresponding texture to be sent to the compositing fshader.
        Typically a filter with internal stages wants its final internal texture to be
        provided into the compositing fshader.

        There is no need to override this in derived classes.

        Parameters:

            Texture names.

        Example:

            self.createInternalTextures( "mycooltex0", "mycooltex1" )

        """
        for texName in args:
            self.textures[texName] = makeFilterTexture(texName)


    def update(self, task=None):
        """Update task.

        Override this in derived classes to update those run-time parameters (shader inputs)
        that must be updated every frame. If your filter does not have any such parameters,
        then there is no need to provide this function.

        This is added to taskMgr by FilterStage at the end of reconfigure(), if self.registerUpdatable()
        has been called in setup(). If not, this method is ignored.

        Usual Panda task manager semantics apply; you should return task.cont to keep the task running.

        Cleanup of tasks is done automatically by detach().

        The default implementation of this method simply shuts down the task (if called by the task manager).

        """
        if task is not None:
            return task.done

##################################################################
# Example filters
##################################################################

# The more complex filters will be in their own modules; short, simple
# filters will be collected into a MiscFilters module.

class StageInitializationFilter(Filter):
    """This filter simply initializes pixcolor from the current pixel in the pipeline stage input color texture.

    While it gets the pixcolor parameter as an input, it is not initialized yet, and is not used.

    It is automatically inserted into each FilterStage created during FilterPipeline.reconfigure().

    Unlike other filters, the properties of StageInitializationFilter are not meaningful,
    and the filter is owned by the FilterStage it resides in. Hence, during each reconfigure of the
    pipeline, old StageInitializationFilters are destroyed and new ones created (because FilterStages are).

    The sort value must be -1, which places StageInitializationFilter before any other filters in the stage.
    Only StageInitializationFilter is allowed to have a sort value of -1.

    """
    def __init__(self, pipeline, **kwargs):
        super(self.__class__, self).__init__(pipeline, **kwargs)

    def reset(self):
        super(self.__class__, self).reset()  # reset inherited properties
        self.sort        = -1     # Initialization goes first before any other filter
                                  # in the same FilterStage.

        self.stageName   = None   # This filter is special; the value of stageName is not used.
                                  # Instead, each created render pass gets one copy of the filter.

        self.isMergeable = None   # The initialization filter is only added after the logical stage
                                  # into render pass merging process is complete. The value of isMergeable
                                  # is ignored.

    def setup(self, phase):
        if phase == "pipeline":
            # The compositing shader needs the color texture to initialize the output color.
            self.registerInputTexture(texName="color", needTexpix=False)

    def synthesize(self):
        # Registered textures are accessed with the prefix "k_tx"; e.g. "color" is "k_txcolor" in the code.
        #
        # self.filterStage.texcoord(texName) retrieves the name of the texcoord variable
        # currently assigned for the texture texName (e.g. "color").
        #
        # pixcolor is the input/output variable for the color of the current pixel.
        #
        texcoordC = self.filterStage.texcoord("color")
        return ("initializeFilterStage",
                "pixcolor = tex2D(k_txcolor, %s.xy);\n" % texcoordC,
                "// initialize pixcolor\n")


# A simple filter. Examples of:
#  - Storing filter parameters in properties
#  - Compile-time parameter
#
class TintFilter(Filter):
    """Color tint / fade-to-color filter.

    In tint mode [default]:

      The color of each pixel is tinted (multiplied) by the specified constant color.
      The result is saturated to [0,1] in each component.

      The tinted pixel then is blended with the original one with the strength
      specified by blend.

    In fade mode:

      Each pixel is blended toward the specified constant color.
      The parameter blend controls the strength.

    This filter can be used e.g. to fade the screen to black (color=(0,0,0,1))
    by dynamically updating the blend parameter in your own update task
    so that it gradually goes from 0 to 1.

    The fade mode also allows fading toward colors other than black
    (e.g. color=(1,1,1,1) to fade to white).

    In tint mode, fading the color to (255,255,255,1) gives another kind of
    fade to white (using the fact that there are 8 bits per color channel).

    Parameters:

      mode  = string, one of:
                "tint" = tint (multiply) by the specified color.
                "fade" = fade to the specified color.

      color = (R,G,B,A) tuple; tint color. To make a component no-op in tint mode,
              set it to 1. Usually setting the alpha component to 1 is a good idea
              regardless of the mode.

      blend = strength of blending between the original and processed color.
              1.0 means fully tinted or faded, 0.0 does nothing.
              Behaves linearly.

    """
    def __init__(self, pipeline, **kwargs):
        super(self.__class__, self).__init__(pipeline, **kwargs)

    def reset(self):
        super(self.__class__, self).reset()  # reset inherited properties
        self.sort  = 90  # late within the pipeline stage
        self.stageName = "Postprocess"
        self.color = (0.5,0.5,0.5,1)  # default is darken by 50%
        self.blend = 1.0
        self.mode  = "tint"  # compile-time parameter

    def setup(self, phase):
        if phase == "pipeline":
            # We do not need to register the "color" texture, as our code does not access it explicitly,
            # and we do not need texture coordinates, either.
            #
            # The StageInitializationFilter sets up the initial value of pixcolor automatically.
            # It is thus better to refer to pixcolor for the original color of the pixel being processed,
            # to save one texture lookup (and two unnecessary parameters to our Cg function generated
            # in synthesize()).
            #
            self.registerCustomInput(inputType="float4", inputName="k_tint_color")
            self.registerCustomInput(inputType="float",  inputName="k_tint_blend")

    def synthesize(self):
        code = ""
        if self.mode == "tint":
            code += "//   Mode: color tint\n"
            code += "float4 tinted_color = saturate(pixcolor * k_tint_color);\n"
            code += "pixcolor = lerp(pixcolor, tinted_color, k_tint_blend);\n"
        elif self.mode == "fade":
            code += "//   Mode: fade-to-color\n"
            code += "pixcolor = lerp(pixcolor, k_tint_color, k_tint_blend);\n"
        else:
            raise ValueError("TintFilter: invalid mode '%s'; valid: 'tint', 'fade'" % self.mode)
        return ("tintFilter", code, "// Color tint / fade-to-color filter\n")

    color = Filter.makeRTParam("_color", "Tint color as (R,G,B,A) float tuple, each component in [0,1].",
                               "finalQuad", "tint_color")

    blend = Filter.makeRTParam("_blend", "Effect strength, float. 1 is full strength, 0 is off.",
                               "finalQuad", "tint_blend")

    mode  = Filter.makeCTParam("_mode", "Effect mode, either 'tint' or 'fade'.")


# A filter that needs texpix.
#
class ScanlinesFilter(Filter):
    FIELD_TOP    = 0
    FIELD_BOTTOM = 1

    def __init__(self, pipeline, **kwargs):
        super(self.__class__, self).__init__(pipeline, **kwargs)

    def reset(self):
        super(self.__class__, self).reset()
        self.sort  = 95  # should probably be applied last
        self.stageName = "DisplayDevice"
        self.field = ScanlinesFilter.FIELD_TOP
        self.blend = 0.5
        self.line_thickness = 1
        self.dynamic = False

    def setup(self, phase):
        if phase == "pipeline":
            # This filter needs texpix and texture coordinate for the color texture,
            # so we must register the texture, even though we will not access k_txcolor directly.
            #
            # (This costs us one unnecessary parameter in the parameter list for scanlinesFilter(),
            #  because the unused k_txcolor is passed in. Currently it is not possible to get
            #  just the auxiliary variables without the actual texture.)
            #
            self.registerInputTexture(texName="color",  needTexpix=True)
            self.registerCustomInput(inputType="int",   inputName="k_scanlines_field")
            self.registerCustomInput(inputType="float", inputName="k_scanlines_blend")
            self.registerCustomInput(inputType="int",   inputName="k_scanlines_thickness")

            # This filter, when in dynamic mode, requires updating the "field" input at each frame.
            #
            # We always add the task so that "dynamic" will behave as expected if switched on/off
            # on the fly.
            #
            self.registerUpdatable()

    def synthesize(self):
        texcoordC = self.filterStage.texcoord("color")
        code  = "if (int(%s.y / texpix_txcolor.y)/k_scanlines_thickness % 2 != k_scanlines_field)\n" % texcoordC
        code += "  pixcolor.rgb = (1.0 - k_scanlines_blend) * pixcolor.rgb;\n"
        return ("scanlinesFilter", code, "// CRT TV scanlines filter\n")

    def update(self, task=None):
        # Note that we use the value of an UTParam (meant for update() to read)
        # to decide whether to update an RTParam (shader input).
        #
        if self.dynamic:
            self.field = 1 - self.field
        if task is not None:
            return task.cont

    field = Filter.makeRTParam("_field", "Which field to keep, either ScanlinesFilter.FIELD_TOP or ScanlinesFilter.FIELD_BOTTOM.",
                               "finalQuad", "scanlines_field")

    blend = Filter.makeRTParam("_blend", "Effect strength (float). 1 is fully scanlined, 0 does nothing.",
                               "finalQuad", "scanlines_blend")

    line_thickness = Filter.makeRTParam("_line_thickness", "Thickness of each scanline in pixels (integer).",
                                        "finalQuad", "scanlines_thickness")

    dynamic = Filter.makeUTParam("_dynamic", "Dynamic mode (bool). When enabled, the 'field' parameter is toggled at each frame rendered. With a steady refresh rate synced to display refresh, this creates an old CRT TV appearance, but the effect is very sensitive to even small unevenness in the framerate. (Hence, default False.)")


# A filter with internal stages.
#
class BlurSharpenFilter(Filter):
    """A blur/sharpen filter. If the 'amount' parameter is 1.0, it will not have effect.
    A value of 0.0 means fully blurred, and a value higher than 1.0 sharpens the image."""

    def __init__(self, pipeline, **kwargs):
        super(self.__class__, self).__init__(pipeline, **kwargs)

    def reset(self):
        super(self.__class__, self).reset()
        self.sort = 0  # first within the pipeline stage (right after init)
                       # (the blur filter inherently cannot respect previous modifications to pixcolor)

        # The input texture for the blur filter must see all previous changes;
        # hence this is not mergeable, and must come at an appropriate point
        # in the simulated image-forming process.
        #
        self.stageName   = "LensFocus"
        self.isMergeable = False

        self.amount = 0.0  # full strength

    def setup(self, phase):
        if phase == "pipeline":
            # Our compositing fshader wants to blend between the original color texture provided
            # by FilterStage, and the internal blur1 texture set up here during FilterStage attach
            # (i.e. the output texture of the last internal stage).
            #
            # Register the "blur1" texture as an input texture to the compositing fshader
            # (in the FilterStage where this filter happens to be assigned), so that the code
            # provided in synthesize() will have access to it.
            #
            # We do not need to register the "color" texture, as our code does not access it explicitly,
            # and does not need its texture coordinate, either. From the color texture, we only need the
            # color of the current pixel, which is provided in pixcolor.
            #
            self.registerInputTexture(texName="blur1",  needTexpix=False)

            # Our compositing fshader also has a custom input to control the strength of the blur effect.
            #
            # Note that registration only requests FilterStage to write the code to pass in this input;
            # we still need to create a run-time parameter and connect it to shader input "blur_amount"
            # (stripping the "k_" prefix).
            #
            self.registerCustomInput(inputType="float", inputName="k_blur_amount")

        elif phase == "stage":
            # Create interQuads for internal stages.
            #
            # Any setup-time parameters for the internal stages are applied here.
            # Usually this means configuring input and output textures to set up an internal pipeline.
            #
            # Note that the output texture for each interQuad is chosen in the call to renderQuadInto().
            # It must be an internal texture; only the compositing fshader can render to the final output.

            # Create output textures for internal stages.
            #
            self.createInternalTextures( "blur0", "blur1" )

            # Create a quad that outputs to texture "blur0". Connect its shader input named "src"
            # to the input color texture provided to us by the FilterStage this filter is attached to.
            #
            # It doesn't matter that the shader hasn't been created yet. Shader inputs are NodePath attributes,
            # so when we assign a shader later, the input named "src" will magically become available.
            # (It is then up to the shader code to actually use the input having that name.)
            #
            # Note that shader inputs for internal shaders are not registered to FilterStage;
            # the registration mechanism only controls code generation of the compositing fshader.)
            #
            # We use div=2 to make a quarter-resolution quad relative to the resolution of the input texture.
            # This is good enough for blur, and usually renders much faster due to much fewer fshader calls.
            #
            self.interQuads.append(self.pipeline.manager.renderQuadInto(colortex=self.textures["blur0"],div=2))
            self.interQuads[0].setShaderInput("src", self.inputTextures["color"])

            # Create a quad that outputs to texture "blur1". Connect its shader input named "src"
            # to the texture "blur0" (which is the output texture of the previous step).
            #
            # Now there is no div=..., so this quad outputs at its input resolution. The input "blur0"
            # is already quarter-resolution; hence the output "blur1" will be, too.
            #
            self.interQuads.append(self.pipeline.manager.renderQuadInto(colortex=self.textures["blur1"]))
            self.interQuads[1].setShaderInput("src", self.textures["blur0"])

    def make(self):
        # Compile and assign shaders for internal stages.
        #
        # These particular shaders have no compile-time parameters, but if they did,
        # those would be applied here by generating the shader code in the appropriate
        # manner based on their current values (the values can be obtained by simply
        # reading the properties corresponding to the appropriate Filter-level
        # compile-time parameters).
        #
        self.interQuads[0].setShader(self.loadShader("filter-blurx.sha"))
        self.interQuads[1].setShader(self.loadShader("filter-blury.sha"))

    def synthesize(self):
        texcoordB = self.filterStage.texcoord("blur1")
        code = "pixcolor = lerp(tex2D(k_txblur1, %s.xy), pixcolor, k_blur_amount.x);\n" % texcoordB
        return ("blurSharpenFilter", code, "// Blur/sharpen blend pass\n")

    # Create a run-time parameter connected to the shader input "blur_amount",
    # corresponding to the variable "k_blur_amount" in the shader code.
    #
    # Note that this definition belongs in the class scope.
    #
    # Parameters are implemented as Python properties; makeRTParam() implements some template code
    # and returns the created property.
    #
    # There is no type checking of the input at this level; the value assigned to the property
    # (in code using this filter) is passed to setShaderInput().
    #
    amount = Filter.makeRTParam("_amount", "Effect strength, float. 1 is off; values in [0,1) blur, values in (1,2] sharpen.",
                                 "finalQuad", "blur_amount")


# A filter that needs the glow map in the alpha channel of the input color texture.
#
class ViewGlowFilter(Filter):
    def __init__(self, pipeline, **kwargs):
        super(self.__class__, self).__init__(pipeline, **kwargs)

    def reset(self):
        super(self.__class__, self).reset()
        self.stageName = "DebugHelpers"  # this is a debug helper filter which should go last
        self.sort = 99  # last thing in the stage

    def setup(self, phase):
        if phase == "pipeline":
            self.registerAuxBits(bitmask=AuxBitplaneAttrib.ABOGlow)

    def synthesize(self):
        return ("viewGlowFilter",
                "pixcolor.r = pixcolor.a;\n",
                "// show the glow map by overwriting the red channel (debugging helper filter)\n")


##################################################################
# Filter stage: one render pass in the pipeline
##################################################################

class FilterStage(BatchUpdatable):
    """Single render pass. Several FilterStages together make up a FilterPipeline."""

    def __init__(self, pipeline, stageName):
        """Constructor.

        Parameters:
            pipeline  = FilterPipeline instance where this FilterStage instance belongs to.

            stageName = String. Human-readable name for this FilterStage (used for error messages).

        """
        # batch update mechanism
        super(self.__class__, self).__init__()

        self.filters = []  # must exist before cleanup() is called
        self.cleanup()
        self.pipeline  = pipeline  # FilterPipeline instance
        self.stageName = stageName # human-readable string; identifies the stage (for error messages)


    def __del__(self):
        """Destructor."""
        self.cleanup()


    def ls(self, indent=0):
        """Print a human-readable description of this FilterStage instance into the terminal."""
        ind = (indent * " ")
        nf  = len(self.filters)
        hpsStatus = ", HalfPixelShift enabled" if self.halfPixelShift else ""

        if nf < 1:
            print("%sFilterStage instance '%s' at 0x%x: <no filters%s>" % (ind, self.stageName, id(self), hpsStatus))
        else:
            # Even with only HalfPixelShift, there is always the StageInitializationFilter.

            fplural = "s" if nf != 1 else ""
            print("%sFilterStage instance '%s' at 0x%x: <%d filter%s%s>" % (ind, self.stageName, id(self), nf, fplural, hpsStatus))

            if len(self.consolidatedTextures) > 0:
                # Find out which filters enabled which textures
                # (in general, several filters may enable the same texture)
                #
                texToFilters = {}
                for f in self.filters:
                    if f not in self.registeredTextures:
                        continue
                    # needTexpix is currently not reported in ls().
                    for texName,needTexpix in self.registeredTextures[f]:
                        texToFilters[texName].append( f )

                # sort by texture name for human-readability; add list of filters that registered each texture
                textureslist = []
                for texName in sorted(self.consolidatedTextures.keys()):
                    filternames = [ f.__class__.__name__ for f in texToFilters[texName] ]
                    # printing the list will already add [...], we don't want to bracketize twice
                    textureslist.append( "%s (reg. by %s)" % (texName, filternames) )

                print("%s  Textures registered to compositing shader: %s" % (ind, textureslist) )

            if len(self.consolidatedCustomInputs) > 0:
                # These are uniquely named, so each custom input corresponds only to one filter.
                #
                inputsByName = {}
                for f in self.filters:
                    if f not in self.registeredCustomInputs:
                        continue:
                    for inputType,inputName in self.registeredCustomInputs[f]:
                        inputsByName[inputName] = "%s %s (reg. by ['%s'])" % (inputType, inputName, f.__class__.__name__)

                # sort by input name for human-readability
                inputslist = []
                for key in sorted(inputsByName.keys()):
                    inputslist.append( inputsByName[key] )

                print("%s  Custom inputs registered to compositing shader: %s" % (ind, inputslist) )

            for f in self.filters:
                f.ls(indent=indent+4)


    def cleanup(self):
        """Cleanup function. Deletes any dynamically created state."""
        # This function must not raise exceptions; it may be called in response to exceptions.

        # Detach the Filters from the FilterStage.
        for f in self.filters:
            f.detach(phase="stage")
        self.filters  = []
        self.pipeline = None

        self.finalQuad = None  # the output quad of this FilterStage (not necessarily of the whole pipeline)
        self.stageInputTextures = {}  # input "scene textures" obtained from the pipeline

        # TODO: BatchUpdatable currently has no reset mechanism. Complicated to provide,
        # TODO: because of possible propagated changes via the hooks.

        # per-filter input parameters
        self.registeredTextures       = {}
        self.registeredCustomInputs   = {}

        # compositing fshader input parameters (computed from per-filter data)
        self.consolidatedTextures     = {}
        self.consolidatedCustomInputs = {}

        # run-time update tasks defined by filters
        self.updateTasks = {}

        # general FilterStage-level parameters
        self._halfPixelShift = False
        self._shaderSourceCode = None
        self._texcoords = {}  # key: texture name, value: texcoord variable name

        # This accounts for FilterStage-level changes requiring recompile.
        #
        # Filters attached to us have (write) access to this to inform us
        # of changed compile-time parameters.
        #
        self._needsCompile = False


    # Note: shaders of internal stages do *not* need to support HalfPixelShift, because *the compositing shader*
    # will apply the shift when reading in the registered textures created by the internal shaders.
    #
    # E.g. the data in the original "color" texture is non-shifted anyway, so any internal textures computed
    # from it should be, too. Then, in the compositing shader, when both "color" and those internal textures
    # are read using the same shift value, we get consistent data.
    #
    @property
    def halfPixelShift(self):
        """Bool. If True, shift the output of this FilterStage by half a pixel in both x and y directions. Default False."""
        # This is actually implemented by shifting the *input* to the compositing shader, but as a description
        # of how it affects the output, the above is correct, and potentially less worrying ("but will it
        # consider my internal stages?").
        #
        return self._halfPixelShift
    @halfPixelShift.setter
    def halfPixelShift(self, value):
        oldValue = self._halfPixelShift
        self._halfPixelShift = value
        if value != oldValue:
            self._needsCompile = True
            self.reconfigure()


    def texcoord(self, texName):
        """Return the name of the texcoord variable corresponding to texture name texName (in the compositing fshader).

        This is a callback for Filter.synthesize().

        Always use the names returned by this function to access texture coordinate variables in code generated
        in Filter.synthesize(); never hard-code the names. (You can assume, though, that the texcoord variable
        name for any given texName remains constant during one invocation of synthesize().)

        Example:

            texcoordC = self.filterStage.texcoord("color")

        returns "l_texcoord" or "l_texcoord_color" depending on circumstances (unpadded texture coordinates
        are shared where possible to reduce the number of fshader arguments; see source code of
        FilterStage.reconfigure() for details).

        """
        assert( len(self._texcoords) > 0 )  # only called via FilterStage.reconfigure()
        return self._texcoords[texName]


    def connect(stageInputTextures, outputQuad):
        """Connect the FilterStage to input and output.

        This assigns to this FilterStage the given input textures and the output quad
        into which the compositing shader (in this FilterStage) renders.

        Parameters:

            stageInputTextures  = Input texture dictionary, containing input textures
                                  for this FilterStage.

                                  May contain entries keyed by names in Filter.SCENETEXTURES,
                                  their values containing the references to the corresponding
                                  Texture instances.

            outputQuad          = NodePath instance; output quad for this FilterStage.

                                  This is the destination where the compositing shader will be
                                  assigned. This will also be passed to any assigned Filters
                                  at FilterStage attach() time.

                                  Note that it is the output quad of *this FilterStage*.
                                  It is not the finalQuad of the whole pipeline,
                                  unless this is the last stage of the pipeline.

        """
        # This is an internal function called only by FilterPipeline.reconfigure().
        assert( outputQuad is not None )
        assert( isinstance(outputQuad, NodePath) )
        assert( type(stageInputTextures) is dict )
        # We must not assert() on the number of entries in stageInputTextures, because it is legal
        # for a filter not to register any input textures (although it will be a very boring filter).

        self.stageInputTextures = stageInputTextures
        self.finalQuad          = outputQuad


    def _registerInputTexture(self, filterInstance, texName, needTexpix=False):
        """Register an input texture for a filter.

        This method is intended to be called from FilterPipeline after it has created
        the FilterStages, and knows which Filter ends up in which FilterStage.
        It then propagates the registration information to us.

        For the user callback (for setup() in classes derived from Filter),
        see FilterPipeline._registerInputTexture().

        Registered input textures will be inserted to the parameter list of the compositing fshader
        of the pipeline stage. They will also be inserted to the Cg function call for the filter
        that requested the register.

        """
        assert( filterInstance is not None )
        assert( isinstance(filterInstance, Filter) )

        # Add the texture to per-filter inputs.
        #
        # The per-filter data is used for generating the Cg function call to the user-specified code
        # (inserted into the compositing fshader), and for composing the function definition.
        #
        if filterInstance not in self.registeredTextures:
            self.registeredTextures[filterInstance] = []
        self.registeredTextures[filterInstance].append( (texName, needTexpix) )

        # Add the texture to compositing fshader inputs.
        #
        # This is basically a logical OR of the per-filter inputs.
        #
        # This data is used for generating the parameter list of the compositing fshader.
        #
        if texName in self.consolidatedTextures:
            oldNeedTexpix = self.consolidatedTextures[texName][1]
            needTexpix    = needTexPix or oldNeedTexpix
        self.consolidatedTextures[texName] = (texName, needTexpix)

    # TODO: name: _registerCustomInput() or _registerCustomShaderInput()?
    def _registerCustomInput(self, filterInstance, inputType, inputName):
        """Register a custom shader input for a filter.

        This method is intended to be called from FilterPipeline after it has created
        the FilterStages, and knows which Filter ends up in which FilterStage.
        It then propagates the registration information to us.

        For the user callback (for setup() in classes derived from Filter),
        see FilterPipeline._registerCustomInput().

        Registered custom inputs will be inserted to the parameter list of the compositing fshader
        of the pipeline stage. They will also be inserted to the Cg function call for the filter
        that requested the register.

        """
        assert( filterInstance is not None )
        assert( isinstance(filterInstance, Filter) )

        # TODO: what if multiple filters of the same type? To support that we need name mangling....

        # Catch duplicates before modifying anything.
        #
        item = (inputType,inputName)
        if inputName in consolidatedCustomInputs:
            raise ValueError("In stage '%s': attempting to add custom input '%s'; name is already in use. Existing data for this name: '%s'" % (self.stageName, item, self.consolidatedCustomInputs[inputName]))

        # Add the custom input to per-filter inputs.
        #
        # The per-filter data is used for generating the Cg function call to the user-specified code
        # (inserted into the compositing fshader), and for composing the function definition.
        #
        if filterInstance not in self.registeredCustomInputs:
            self.registeredCustomInputs[filterInstance] = []
        self.registeredCustomInputs[filterInstance].append( item )

        # Add the texture to compositing fshader inputs.
        #
        # This is basically a flattened version of the per-filter inputs.
        # This is guaranteed to be free of duplicates.
        #
        # This data is used for generating the parameter list of the compositing fshader.
        #
        self.consolidatedCustomInputs[inputName] = item

    def _registerUpdatable(self, filterInstance, taskFunction):
        """Register an update task for a filter.

        This method is intended to be called from FilterPipeline after it has created
        the FilterStages, and knows which Filter ends up in which FilterStage.
        It then propagates the registration information to us.

        For the user callback (for setup() in classes derived from Filter),
        see FilterPipeline._registerUpdatable().

        Registered update tasks will be added to the task manager (taskMgr)
        when reconfigure() is otherwise done.

        """
        assert( filterInstance is not None )
        assert( isinstance(filterInstance, Filter) )
        assert( taskFunction is not None )

        self.updateTasks[filterInstance] = taskFunction


    def batchUpdateDone(self):
        """Recompiles this FilterStage when the batch update is done."""
        self.reconfigure()

    # Note:
    #
    # FilterStage does not own the Filter instances, and the official way of doing a batch update
    # is always through FilterPipeline; FilterStage is in a sense an internal class and should not
    # be called into by the user.
    #
    # However, if someone (or some internal mechanism) does happen to start a batch update
    # through FilterStage, the options are either to fail silently (by not hooking; not good),
    # or to hook and propagate the update to any assigned Filters (to make the hierarchy
    # work as expected; on the surface, it appears that Filters are collected into FilterStages,
    # which are collected into a FilterPipeline; the dynamic, volatile nature of FilterStage
    # is not at all apparent).
    #
    # Avoiding the silent failure is preferable; however it has the side effect that
    # when the batch update is requested the official way, via FilterPipeline, the update
    # propagates to Filters twice - once by FilterPipeline and once by the FilterStage where
    # they are assigned.
    #
    # In practice, this doesn't do any harm, because 1) each BatchUpdatable object counts the
    # batch update calls it has received, and 2) even if the calls weren't counted, all the
    # objects in the filter system have the _needsCompile tracking mechanism, which makes
    # any unnecessary reconfigures (which have no changes to apply) to simply no-op.
    #
    # Thus, it is better to hook into the batch updates and propagate them to any assigned Filters.
    #
    def beginBatchUpdateImplHook(self):
        """Propagates the batch update to any Filters currently attached to this FilterStage."""
        for f in self.filters:
            f.beginBatchUpdateImpl()

    def endBatchUpdateImplHook(self, notify):
        """Propagates the batch update to any Filters currently attached to this FilterStage."""
        for f in self.filters:
            f.endBatchUpdateImpl(notify)

    # We also provide the other direction of batch update routing: from FilterStage to FilterPipeline.
    #
    def beginBatchUpdate(self):
        """Begin a batch update, accounting for the filter system hierarchy.

        If self.pipeline is available, begin batch update there.

        Otherwise, begin batch update locally.

        """
        if self.pipeline is not None:
            self.pipeline.beginBatchUpdateImpl()  # this propagates to FilterStage and Filter
        else:
            self.beginBatchUpdateImpl()  # this propagates to Filter

    def endBatchUpdate(self, notify=True):
        """End a batch update, accounting for the filter system hierarchy.

        If self.pipeline is available, end batch update there.

        Otherwise, end batch update locally.

        """
        if self.pipeline is not None:
            self.pipeline.endBatchUpdateImpl(notify)  # this propagates to FilterStage and Filter
        else:
            self.endBatchUpdateImpl(notify)  # this propagates to Filter


    def addFilterInstance(self, f):
        """Assign a Filter instance to this FilterStage.

        The Filter is added to the list of filters in this FilterStage. This flags a recompile
        as needed, but does NOT trigger it.

        Note that FilterPipeline owns and manages the Filter instances. FilterStages are volatile
        in the sense that they are destroyed and created at each reconfigure() of the pipeline.

        Thus FilterStage has no methods to query or remove filters, or to set filter parameters;
        use the methods of FilterPipeline to do that.

        """
        if f is None:
            raise TypeError("f must not be None")
        if not isinstance(f, Filter):
            raise TypeError("f must be an instance of (a subclass of) Filter; got '%s'" % type(f))
        self.filters.append( f )
        self._needsCompile = True


    def reconfigure(self, force=False, allowDelegation=True):
        """Synthesize and compile shaders.

        Usually this works intelligently, only recompiling when needed.

        Parameters:

          force               = bool. If True, recompile regardless of whether this FilterStage instance
                                thinks it is needed or not.

                                No need to use this in user code; used internally between FilterStage and
                                FilterPipeline to support a full pipeline rebuild.

          allowDelegation     = bool. If True, allow delegating the reconfigure to FilterPipeline.

                                No need to use this in user code; used internally between FilterStage and
                                FilterPipeline to support a full pipeline rebuild.

        Both flags are currently needed, because FilterManager has only a monolithic
        cleanup() that destroys *all* dynamic state (i.e. it is not possible to destroy only
        the state related to a given FilterStage). Hence, when one FilterStage is reconfigured,
        all others must be, too. (This is subject to change in the future.)

        """
        # The allowDelegation flag helps to implement a split of responsibilities. When a FilterStage-level
        # compile-time parameter in an individual Filter instance changes, it requests a recompile from the
        # FilterStage to which that specific Filter instance is currently attached. This is conceptually
        # the correct thing to do.
        #
        # It is then the responsibility of FilterStage.reconfigure() to decide what to do with that request.
        #
        # Currently, the FilterStage then requests the pipeline to be reconfigured, triggering a full rebuild.
        # This is one way to do what the Filter instance requested, but not the only possible way, and is
        # subject to change in the future (if a non-monolithic cleanup is later implemented in FilterManager).

        # This function always works like the "fullrebuild" mode in old CommonFilters (up to 1.8.1).
        #
        # Because configuration resides in filter object properties, and their setters take care of
        # updating the appropriate shader inputs, FilterStage.reconfigure() only has work to do when
        # a compositing shader recompile is needed.

        # During a batch update to filter configuration, do not recompile. (This is not an error situation.)
        #
        if self.batchUpdateInProgress:
            return

        # Unless forced, compile only if something affecting the code generation of this FilterStage
        # has changed.
        #
        if not force  and  not self._needsCompile:
            return

        # In delegation mode, pass the call upstream and "wait for further instructions".
        # FilterPipeline will then call FilterStage.reconfigure() with allowDelegation set to False.
        #
        # FilterPipeline.reconfigure() is, and MUST be, the only caller where allowDelegation is False.
        #
        # This must be placed *after* the stage-level checks. Keep in mind that by calling us, the caller
        # essentially wants to recompile this FilterStage only. Hence, if we have nothing to do,
        # we shouldn't delegate the call. (Once we delegate, a full pipeline rebuild begins immediately.)
        #
        # Note that textures and quads are completely destroyed and re-created during each rebuild.
        # This is because of the monolithic cleanup in FilterManager.
        #
        #
        # (Simply not re-creating the textures at each reconfigure would bring another problem:
        #  some filters, e.g. CartoonInk, may require different combinations of input textures
        #  depending on their compile-time options. It has a depth-enabled flag, which makes it
        #  use scene-depth if the flag is enabled, and not even request it if not. If no other filter
        #  happens to need depth, enabling/disabling this setting requires a full cleanup and reconfigure
        #  of FilterManager.
        #
        #  Also, nothing prevents a filter from requiring a different set of internal textures
        #  (and different internal pipelining!) depending on its operating mode; e.g. a blur that
        #  may apply one or two passes depending on settings.
        #
        #  The full rebuild - which was also the solution used in old CommonFilters - is the simplest way
        #  to support such arbitrary configuration changes robustly.)
        #
        #
        # Note that once FilterPipeline.reconfigure() begins, any old FilterStage instances
        # (including this one) will get destroyed and new ones created.
        #
        if allowDelegation:
            # We force, so that the pipeline recompiles even if:
            #  - no pipeline-level parameter has changed
            #  - no filters have been added/deleted
            self.pipeline.reconfigure(force=True)
            return  # the old FilterStage is done, and has actually already been cleaned up for deletion.

        ######################################################################
        # Currently, when we reach this point, a full rebuild is in progress,
        # and we are a new FilterStage in a clean state.
        ######################################################################

        # - We are being called from FilterPipeline.reconfigure(),
        #   so we can consider ourselves an internal function. Hence asserts.
        #
        # - connect() must have been called first to set up the output quad
        #   and input scene textures, before it is valid to attach() filters
        #   to the FilterStage. Check that now before we begin the actual build.
        #
        assert( self.finalQuad is not None )
        assert( isinstance(self.finalQuad, NodePath) )
        assert( type(self.stageInputTextures) is dict )
        # We must not assert() on the number of entries in self.stageInputTextures, because it is legal
        # for a filter not to register any input textures (although it will be a very boring filter).

        # Also sanity-check the enabled filters.
        #
        # addFilterInstance() already does this, but the data member is public (TODO: maybe shouldn't be?)
        #
        for f in self.filters:
            assert( f is not None )
            assert( isinstance(f, Filter) )

        # Validate sort values for all filters, and if the check passes, then sort the filters.
        #
        # The check can fail due to errors in user-defined code, so here we raise an exception on error.
        #
        sortvalues = {}  # for sort conflict error message
        for f in self.filters:

            # definedness
            if f.sort is None:
                raise ValueError("In FilterStage '%s': filter %s has no sort value (sort is None). Valid: integer, 0 <= sort < 100." % (self.stageName, f.__class__.__name__))

            # range
            if isinstance(f, StageInitializationFilter)  and  f.sort != -1:
                raise ValueError("In FilterStage '%s': stage initialization filter %s must have sort = -1, but has %d." % (self.stageName, f.__class__.__name__, f.sort))
            elif f.sort < 0  or  f.sort > 99:
                raise ValueError("In FilterStage '%s': filter %s has out-of-range sort value %d. Valid: integer, 0 <= sort < 100." % (self.stageName, f.__class__.__name__, f.sort))

            # uniqueness (better to raise an error than let containers decide what to do in case of conflict;
            #             the whole idea of the sort value is that it defines an explicit ordering for the
            #             filters.)
            #
            if f.sort in sortvalues:
                raise ValueError("In FilterStage '%s': sort conflict between filters %s and %s; both have sort value %d." % (self.stageName, f.__class__.__name__, sortvalues[f.sort].__class__.__name__, f.sort))
            sortvalues[f.sort] = f  # remember which filter had this sort value (for conflict error message)

        # This accounts also for stage mergings by FilterPipeline. At first, f.mergings is zero
        # for all filters "f", and each stage merge for a filter "f" in the current rebuild
        # increases f.mergings by one.
        #
        self.filters.sort(key=lambda f: 100*f.mergings + f.sort)

        # Attach filters to the stage.
        #
        # Note that the input textures MUST be available before this is done; the filters
        # store their stage input texture references from us at FilterStage attach time.
        #
        # attach() calls setup() to create internal stages if the filter has any,
        # and (now that the interQuads are available) recompiles the filter's internal shaders
        # if it has any.
        #
        for f in self.filters:
            f.attach(filterStage=self)

        # Gather Texture instance references into one place, regardless whether they are scene textures
        # (FilterStage input) or internal textures defined by filters.
        #
        allRegisteredTextures = {}  # key: texture name, value: Texture instance reference
        neededTextures = self.consolidatedTextures.keys()  # textures needed in compositing shader; names only
        neededSceneTextures = filter( lambda s: s in Filter.SCENETEXTURES, neededTextures )
        for texName in neededSceneTextures:
            allRegisteredTextures[texname] = self.stageInputTextures[texName]
        # internal textures defined by the filters
        for f in self.filters:
            for texName in f.textures.keys():  # f.textures will be blank for most filters (no internal stages)
                if texName in neededTextures:  # this internal texture needed in compositing shader
                    allRegisteredTextures[texname] = f.textures[texName]

        # Set texture inputs for our compositing shader.
        #
        for texName,ref in allRegisteredTextures:
            self.finalQuad.setShaderInput("tx"+texName, ref)

        # Figure out which texcoord variables (for both vshader and fshader),
        # and which texpad and texpix variables (for vshader only), will be needed.
        #
        # (The fshader does not need texpad, and which texpix it needs depends on
        #  the needTexpix flags registered by the filters.)
        #
        texcoords       = {}
        texcoordPadding = {}
        for texName in allRegisteredTextures.keys():
            # texpad and texpix (the latter needed in the vshader by HalfPixelShift)
            # must be handled per-texture due to possibly different texture sizes.
            #
            # texpad holds the coordinates of the center of the texture, in texture coordinates.
            # If this is a "textures-power-2 none" texture (i.e. not padded), then the center position
            # is always (0.5, 0.5), and we can omit that from the shader inputs.
            #
            # HalfPixelShift implies per-texture texcoord due to possibly different texture sizes,
            # i.e. different textures will require different amounts of shift in texture coordinates.
            #
            # Most of the time, all filter textures are "textures-power-2 none", and HalfPixelShift
            # is not enabled, allowing us to provide only one shared texcoord variable.
            #
            texPadded = (allRegisteredTextures[texName].getAutoTextureScale() != ATSNone)
            if texPadded  or  self.halfPixelShift:  # need per-texture texcoord?
                texcoords[texName] = "l_texcoord_" + texName
                padName = texName if texPadded else None
                pixName = texName if self.halfPixelShift else None
                texcoordPadding["l_texcoord_" + texName] = (padName, pixName)
            else:
                # Share unpadded texture coordinates.
                texcoords[texName] = "l_texcoord"
                texcoordPadding["l_texcoord"] = (None, None)
        self._texcoords = texcoords
 
        # number for matching TEXCOORD registers between vshader and fshader,
        # name for the corresponding texcoord variable.
        texcoordSets = list(enumerate(texcoordPadding.keys()))


        ########################################################
        # Synthesize compositing shader for this pipeline stage
        ########################################################

        # complete program = language header + vertex shader + filter functions + fragment shader

        # top-level indentation depth, in spaces
        #
        tabWidth = 4

        # We use the ancient arbvp1 and arbfp1 profiles, because when using any newer profile supported by Cg,
        # the shaders compiled by Cg only work properly on NVIDIA cards. Shaders compiled using arbvp1/arbfp1
        # profiles work properly on all cards, at least those supporting shader model 2.0 and above.
        #
        # The good thing is that arbvp1/arbfp1 shaders will run on even rather ancient (2007) hardware.
        #
        # TODO: Port this and all Filters over to GLSL, since Cg is no longer maintained.
        # TODO: Decide later what to do about supporting old hardware at that point.

        cgHeader  = "//Cg\n"
        cgHeader += "//\n"
        cgHeader += "//Cg profile arbvp1 arbfp1\n\n"

        ##################
        # compose vshader
        ##################

        vshader  = "void vshader( "
        ind = len(vshader)  # amount of indentation in parameter list
        vshader += "float4 vtx_position : POSITION,\n"
        vshader += "%sout float4 l_position : POSITION,\n" % ind

        for padTex,pixTex in texcoordPadding.values():
            if padTex is not None:
                vshader += "%suniform float4 texpad_tx%s,\n" % (ind, padTex)
            if pixTex is not None:
                vshader += "%suniform float4 texpix_tx%s,\n" % (ind, pixTex)
        for i, name in texcoordSets:
            vshader += "%sout float2 %s : TEXCOORD%d,\n" % (ind, name, i)

        vshader += "%suniform float4x4 mat_modelproj )\n" % ind
        vshader += "{\n"
        ind = (tabwidth * " ")  # amount of indentation in vshader body
        vshader += "%sl_position = mul(mat_modelproj, vtx_position);\n" % ind

        for texcoord, item in texcoordPadding.items():
            padTex,pixTex = item
            if padTex is not None:
                vshader += "%s%s = (vtx_position.xz * texpad_tx%s.xy) + texpad_tx%s.xy;\n" % (ind, texcoord,
                                                                                              padTex, padTex)
            else:
                vshader += "%s%s = (vtx_position.xz * float2(0.5, 0.5)) + float2(0.5, 0.5);\n" % (ind, texcoord)
            if pixTex is not None:
                vshader += "%s%s += texpix_tx%s.xy * 0.5;\n" % (ind, texcoord, pixTex)

        vshader += "}\n\n"

        ###########################
        # compose filter functions
        ###########################

        # We build both the functions and their invocations here.

        filterFunctions = ""
        filterCalls = {}  # key: Filter instance reference, value: string, function invocation
        for f in self.filters:
            funcName, funcImplementation, funcComment = f.synthesizeFragmentShader()

            code     = ""
            funcCall = ""

            if funcComment is not None:
                code     += funcComment
                funcCall += _indent(funcComment, tabWidth)

            codeStart = "float4 %s( " % funcName
            callStart = _indent("pixcolor = %s( " % (funcName), tabWidth)
            # amount of indentation in parameter list
            indCode = len(codeStart) * " "
            indCall = len(callStart) * " "

            code     += codeStart 
            funcCall += callStart

            for texName,needTexpix in self.registeredTextures[f]:
                code += "uniform sampler2D k_tx%s,\n%s" % (texName, indCode)
                funcCall += "k_tx%s,\n%s" % (texName, indCall)
                if needTexpix:
                    code += "uniform float4 texpix_tx%s,\n%s" % (texName, indCode)
                    funcCall += "texpix_tx%s,\n%s" % (texName, indCall)

            for inputType,inputName in self.registeredCustomInputs[f]:
                code += "uniform %s %s,\n%s" % (inputType, inputName, indCode)
                funcCall += "%s,\n%s" % (inputName, indCall)

            code += "float4 pixcolor )\n"
            code += "{\n"
            code += _indent(funcImplementation,   tabWidth)
            code += _indent("return pixcolor;\n", tabwidth)
            code += "}\n\n"
            filterFunctions += code

            funcCall += "pixcolor );\n\n"
            filterCalls[f] = funcCall

        ##################
        # compose fshader
        ##################

        # The implementation of the fshader simply consists of calls to enabled filters,
        # ordered by the sort order already determined further above.
        #
        # We use a temporary "pixcolor" variable to prevent implicit saturation to [0, 1]
        # in intermediate processing; assigning to o_color directly would do that.
        #
        # The filters are expected to saturate explicitly, if they want to.
        #
        # We only assign the result to o_color once all the processing in this stage has completed.
        #
        # Initialization of pixcolor (from the current pixel in k_txcolor) is provided
        # by StageInitializationFilter, which always goes first.

        fshader = "void fshader( "
        ind = len(fshader)  # amount of indentation for parameter list

        for i, name in texcoordSets:
            fshader += "float2 %s : TEXCOORD%d,\n%s" % (name, i, ind)

        # For the fshader, we use the needTexpix flags registered by the filters.
        for texName,needTexpix in self.consolidatedTextures:
            fshader += "uniform sampler2D k_tx%s,\n%s" % (texName, ind)
            if needTexpix:
                fshader += "uniform float4 texpix_tx%s,\n%s" % (texName, ind)

        for inputType,inputName in self.consolidatedCustomInputs:
            fshader += "uniform %s %s,\n%s" % (inputType, inputName, ind)

        fshader += "out float4 o_color : COLOR )\n"
        fshader += "{\n"
        fshader += _indent("float4 pixcolor;\n", tabWidth)
        for f in self.filters:
            fshader += filterCalls[f]
        fshader += _indent("o_color = pixcolor;\n", tabWidth)
        fshader += "}\n"

        ########################################################

        # Assemble the complete shader program, compile the shader,
        # and assign it to our output quad.
        #
        # We save the current shader source code into a member variable
        # to aid debugging and run-time inspectability.
        #
        completeSourceCode = cgHeader + vshader + filterFunctions + fshader

        shaderObj = Shader.make(completeSourceCode)
        if shaderObj is None:
            filterlist = [ obj.__class__.__name__ for obj in self.filters ]
            self.cleanup()
            raise RuntimeError("Shader.make() for compositing shader failed in FilterStage '%s', which had the following filters: %s" % (self.stageName, filterlist))
        self.finalQuad.setShader(shaderObj)  # this will keep the Shader reference, so we won't need to
        self._shaderSourceCode = completeSourceCode

        # Start update tasks that were registered by the filters.
        #
        for f,taskFunction in self.updateTasks.items():
            # Although we are FilterStage, and the actual tasks are contained in individual Filters,
            # we put "filterpipeline" into the task name to announce in a human-readable manner
            # which subsystem the task belongs to.
            #
            # TODO: rethink case with multiple instances of the same filter: mark instance somehow
            #
            taskName = "filterpipeline-update-%s" % f.__class__.__name__

            # TODO: what priority value to use to make this run just before render?
            f.task = taskMgr.add(taskFunction, taskName)

            # Make the task run once immediately.  TODO: the old version did this, but do we really need this?
            taskFunction()

        # The compile of this FilterStage is now complete.
        #
        self._needsCompile = False

    @property
    def shaderSourceCode(self):
        """Source code of the currently active compositing shader of this FilterStage, or None if not running.

        Read-only property.

        Note that the source may contained compiled-in values for filter parameters,
        especially for loop limits (to conform to the arbvp1 and arbfp1 profiles,
        which do not allow variable loop limits; these profiles require all loops
        to be unrolled by the compiler, and hence loop limits must be known at compile time).

        Also, the exact content of the source code will (obviously) depend on which filters
        are present in this FilterStage, and on their options. Note especially that
        filters are allowed to provide options that completely change their generated code
        (e.g. an algorithm selection option).

        Thus, this is mainly useful for debugging and examining the framework; for understanding
        individual Filter subclasses, it may be clearer to read their sources directly.

        """
        return self._shaderSourceCode


######################################################
# Pipeline: manages filters and render passes
######################################################

class FilterPipeline(BatchUpdatable):
    """FilterPipeline implements the postprocessing filter pipeline.

    This class provides the main part of the user API for the postprocessing filter system.

    See methods:

      setFilter()    (automatic add/reconfigure)
      delFilter()

    Alternative ways to add/reconfigure filters:

      addFilter()
      addFilterInstance()
      setFilterParameters()

    Also interesting:

      findFilter()   (to access the Filter objects directly)


    Internally, the main responsibilities of this class are to manage Filter instances and
    render passes (FilterStage instances), and to dynamically construct actual render passes
    from the logical stages (Filter.STAGES) by examining the configuration of enabled filters.
    Also the fullscreen textures and quads are managed by FilterPipeline (with the help of
    FilterManager for the low-level work).

    """
    def __init__(self, win, cam):
        """Constructor.

        Parameters:

          win = Window to postprocess.

          cam = Camera that is currently either rendering "win", or rendering some display region of "win"
                (for windows having multiple DisplayRegions). FilterManager will automatically find the
                DisplayRegion associated with the specified camera.

        These are passed as-is to FilterManager; see its documentation for a detailed explanation.

        Example:

          To apply postprocessing filters to the main window in a typical application using DirectStart,
          do this somewhere in your class inheriting from DirectObject:

            self.fpp = FilterPipeline(base.win, base.cam)

          Storing the FilterPipeline instance in a member of self guarantees that it will stay alive
          when your setup method returns.

          If you need to, you can later delete the pipeline by simply

            del self.fpp

          and the window should return to its original state.

        """
        # batch update mechanism
        super(self.__class__, self).__init__()

        # We use FilterManager to manage the actual render-to-quads and their textures.
        #
        self.manager = FilterManager(win, cam)

        # To FilterPipeline, which owns and manages the Filter instances, the enabled
        # filters are an unordered collection. The concept of ordering of filters
        # within a render pass is implemented in FilterStage.
        #
        self.filters = set()

        # FilterPipeline is responsible for the ordering of render passes;
        # hence this needs an ordered collection (such as a list).
        #
        self.stages  = []

        self._halfPixelShift = False

        self.cleanup()


    def __del__(self):
        """Destructor."""
        self.cleanup()  # this calls self.manager.cleanup(), too
        self.filters = set()
        self.manager = None


    @property
    def halfPixelShift(self):
        """Bool. If True, shift the output of the first FilterStage by half a pixel in both x and y directions. Default False."""
        return self._halfPixelShift
    @halfPixelShift.setter
    def halfPixelShift(self, value):
        self._halfPixelShift = value
        # This actually affects the first FilterStage, so pass the modification through.
        #
        # If we don't have any FilterStages at the moment, we are done (in that case, our reconfigure()
        # will eventually apply this when it creates the FilterStages).
        #
        if len(self.stages) >= 1:
            self.stages[0].halfPixelShift = value  # The setter will trigger reconfigure for the FilterStage.


    def ls(self, indent=0):
        """Print a human-readable description of this FilterPipeline instance into the terminal."""
        ind = (indent * " ")
        nrp = len(self.stages)
        if nrp < 1:
            print("%sFilterPipeline instance at 0x%x: <inactive>" % (ind, id(self)))
        else:
            hpsStatus = ", HalfPixelShift enabled" if self.halfPixelShift else ""
            rpplural = "es" if nrp != 1 else ""
            nf  = len(self.filters)
            fplural  = "s"  if nf  != 1 else ""
            print("%sFilterPipeline instance at 0x%x: <active>, %d render pass%s, %d filter%s total%s" % (ind, id(self), nrp, rpplural, nf, fplural, hpsStatus))
            # sort by texture name for human-readability
            print("%s  Scene textures: %s" % (ind, sorted(self.sceneTextures.keys())) )
            # stages have an ordering; present them in their correct order.
            for i,stage in enumerate(self.stages):
                print("%s  Render pass %d/%d:" % (ind,i+1,nrp))
                stage.ls(indent=indent+4)


    def cleanup(self):
        """Cleanup function.

        Destroys dynamic state, leaving the FilterPipeline into a state suitable to begin a rebuild.

        Does NOT destroy Filter instances owned by this FilterPipeline; those are only destroyed
        when explicitly asked to do so (delFilter()), or when the FilterPipeline is destroyed.
        This makes the filters preserve their configuration across pipeline reconfigures.

        This is intended to be called from reconfigure().

        """
        # This function must not raise exceptions; it may be called in response to exceptions.

        # per-filter input parameters (passed through to FilterStages once we know which filter goes where)
        self.registeredTextures       = {}
        self.registeredCustomInputs   = {}

        # compositing fshader input parameters (computed from per-filter data)
        self.consolidatedTextures     = {}
        self.consolidatedCustomInputs = {}

        # general behavioral parameters (computed from per-filter data)
        self.consolidatedAuxBits = 0

        # run-time update tasks defined by filters
        self.updateTasks = {}

        # Clean up and destroy the current FilterStages.
        #
        # This detaches the Filters from the FilterStages, and releases the Filter references
        # from the FilterStages. The Filter instances stay owned by this FilterPipeline.
        #
        for s in self.stages:
            s.cleanup()
        self.stages = []

        # Detach the filters from the pipeline. In the cleaned-up state, FilterPipeline does not want
        # batch update or reconfigure callbacks from parameter changes in the filters.
        #
        # This also resets the mergings count of the filters.
        #
        for f in self.filters:
            f.detach(phase="pipeline")

        # Clear input/output texture and quad references
        #
        self.sceneTextures = {}
        self.stageColorTextures = []
        self.quads = []

        # Now that the Filters are no longer attached, destroy textures and quads.
        #
        self.manager.cleanup()

        # All filters in detached state, no need to compile.
        self._needsCompile = False


    def _registerInputTexture(self, filterInstance, texName, needTexpix=False):
        """Register an input texture for a filter.

        Callback. This is meant to be called from setup() in classes derived from Filter,
        but usually via Filter.registerInputTexture(), which sets filterInstance automatically.
        (Hence, hidden method.)

        Registered input textures will be propagated to FilterStage._registerInputTexture()
        once the FilterStages are created during the rebuild.

        FilterPipeline itself uses this information to decide which scene textures are needed.

        """
        assert( filterInstance is not None )
        assert( isinstance(filterInstance, Filter) )

        # Add the texture to per-filter inputs.
        #
        # The per-filter data is used for generating the Cg function call to the user-specified code
        # (inserted into the compositing fshader), and for composing the function definition.
        #
        if filterInstance not in self.registeredTextures:
            self.registeredTextures[filterInstance] = []
        self.registeredTextures[filterInstance].append( (texName, needTexpix) )

        # Add the texture to compositing fshader inputs.
        #
        # This is basically a logical OR of the per-filter inputs.
        #
        # This data is used for generating the parameter list of the compositing fshader.
        #
        if texName in self.consolidatedTextures:
            oldNeedTexpix = self.consolidatedTextures[texName][1]
            needTexpix    = needTexPix or oldNeedTexpix
        self.consolidatedTextures[texName] = (texName, needTexpix)

    # TODO: name: _registerCustomInput() or _registerCustomShaderInput()?
    def _registerCustomInput(self, filterInstance, inputType, inputName):
        """Register a custom shader input for a filter.

        Callback. This is meant to be called from setup() in classes derived from Filter,
        but usually via Filter.registerCustomInput(), which sets filterInstance automatically.
        (Hence, hidden method.)

        Registered custom inputs will be propagated to FilterStage._registerCustomInput()
        once the FilterStages are created during the rebuild.

        """
        assert( filterInstance is not None )
        assert( isinstance(filterInstance, Filter) )

        # TODO: what if multiple filters of the same type? To support that we need name mangling....

        # Catch duplicates before modifying anything.
        #
        item = (inputType,inputName)
        if inputName in consolidatedCustomInputs:
            raise ValueError("Attempting to add custom input '%s'; name is already in use. Existing data: '%s'" % (item, self.consolidatedCustomInputs[inputName]))

        # Add the custom input to per-filter inputs.
        #
        # The per-filter data is used for generating the Cg function call to the user-specified code
        # (inserted into the compositing fshader), and for composing the function definition.
        #
        if filterInstance not in self.registeredCustomInputs:
            self.registeredCustomInputs[filterInstance] = []
        self.registeredCustomInputs[filterInstance].append( item )

        # Add the texture to compositing fshader inputs.
        #
        # This is basically a flattened version of the per-filter inputs.
        # This is guaranteed to be free of duplicates.
        #
        # This data is used for generating the parameter list of the compositing fshader.
        #
        self.consolidatedCustomInputs[inputName] = item

    def _registerAuxBits(self, filterInstance, bitmask):
        """Register an AuxBitplaneAttrib bitmask (e.g. ABOGlow, ABOAuxNormal) for a filter.

        Callback. This is meant to be called from setup() in classes derived from Filter,
        but usually via Filter.registerAuxBits(), which sets filterInstance automatically.
        (Hence, hidden method.)

        Note that FilterStage has no counterpart for this method. FilterPipeline must know
        the aux bits that are needed before it can request the scene textures from the manager.
        The scene textures, in turn, must be set up before the FilterStages can be configured.

        """
        assert( filterInstance is not None )
        assert( isinstance(filterInstance, Filter) )

        # Currently we don't use filterInstance here, as the aux bits only affect global behavior.
        #
        # The parameter is only there for API consistency with the other _register*() methods,
        # to make future expansion easier.
        #
        self.consolidatedAuxBits |= bitmask


    def _registerUpdatable(self, filterInstance, taskFunction):
        """Register an update task for a filter.

        Callback. This is meant to be called from setup() in classes derived from Filter,
        but usually via Filter.registerUpdatable(), which sets the parameters automatically.
        (Hence, hidden method.)

        Registered updatables will be propagated to FilterStage._registerUpdatable()
        once the FilterStages are created during the rebuild.

        """
        assert( filterInstance is not None )
        assert( isinstance(filterInstance, Filter) )
        assert( taskFunction is not None )

        self.updateTasks[filterInstance] = taskFunction


    def batchUpdateDone(self):
        """Recompiles the FilterPipeline when the batch update is done."""
        self.reconfigure()

    def beginBatchUpdateImplHook(self):
        """Propagates the batch update to any FilterStages and Filters in this FilterPipeline."""
        for f in self.filters:
            f.beginBatchUpdateImpl()
        for s in self.stages:
            s.beginBatchUpdateImpl()

    def endBatchUpdateImplHook(self, notify):
        """Propagates the batch update to any FilterStages and Filters in this FilterPipeline."""
        for f in self.filters:
            f.endBatchUpdateImpl(notify)
        for s in self.stages:
            s.endBatchUpdateImpl(notify)


    # TODO: rethink case with multiple instances of the same filter
    def findFilter(self, fclass):
        """Find an enabled filter by its object type (class).

        Parameters:

          fclass  = a class object corresponding to a class derived from Filter.

        Return value:

          (Reference to) Filter instance.

        Example:

          f = myPipeline.findFilter(ScanlinesFilter)

        """
        if not issubclass(fclass, Filter):
            raise TypeError("fclass must be a Filter subclass; got '%s'" % fclass)
        for f in self.filters:
            if isinstance(f, fclass):
                return f
        return None


    # TODO: rethink case with multiple instances of the same filter
    def addFilter(self, fclass, **kwargs):
        """Add (enable) a filter.

        If an instance of the same type of filter is already present, ValueError is raised.


        This is one way to enable filters.

        For an automatic "add or reconfigure" in the style of the old API, see setFilter().

        For adding a manually created Filter instance, see addFilterInstance().


        Parameters:

          fclass = a class object corresponding to a class derived from Filter.

        Kwargs:

          Optional.

          Filter parameters and their values. These are filter-specific.
          Any number of any run-time or compile-time parameters supported
          by filters of type fclass may be given.

          This can be used to override default settings at filter creation time.
          Any parameters not provided will use their default values.

          To update configuration later, see setFilterParameters().

        Return value:

          The created filter instance.

        """
        if not issubclass(fclass, Filter):
            raise TypeError("fclass must be a Filter subclass; got '%s'" % fclass)

        f = self.findFilter(fclass)
        if f is not None:
            raise ValueError("Filter '%s' already present in this pipeline" % fclass)

        self.filters.add( fclass(pipeline=self, **kwargs) )
        self._needsCompile = True
        self.reconfigure()

        return f


    # TODO: rethink case with multiple instances of the same filter
    def delFilter(self, fclass):
        """Remove (disable) a filter.

        There can be only one of each type of filter in the stage,
        so the removal is requested by object type (class).

        If fclass specifies a filter type that is not currently enabled,
        ValueError is raised.

        """
        f = self.findFilter(fclass)

        if f is not None:
            self.filters.remove(f)
            self._needsCompile = True
            self.reconfigure()
        else:
            raise ValueError("Trying to remove filter '%s' that was not present in this pipeline" % fclass)


    # TODO: rethink case with multiple instances of the same filter
    def setFilterParameters(self, fclass, **kwargs):
        """Set parameters to a currently enabled filter (update filter configuration).

        If a filter of type fclass is not currently enabled, ValueError is raised.

        For an automatic "add or reconfigure" in the style of the old API, see setFilter().

        Parameters:

          fclass = a class object corresponding to a class derived from Filter.
                   Identifies which filter the parameters are meant for.

        Kwargs:

          Filter parameters and their new values. These are filter-specific.
          Any number of any run-time or compile-time parameters supported
          by filters of type fclass may be given.

          This is implemented using Filter.updateConfigurationFromKeywords().

          This is an atomic update; the recompile, if needed, runs after
          all specified parameters have been updated.

          If not sure what parameters a given type of filter has, create an
          instance with default settings and query it:

            print MyCoolFilter().getPropertiesWithDocs()

          to get a list of parameters and their docstrings.

        Example:

          filterstage.setFilterParameters(BlurSharpenFilter, amount=0.5)

        """
        if not issubclass(fclass, Filter):
            raise TypeError("fclass must be a Filter subclass; got '%s'" % fclass)

        f = self.findFilter(fclass)
        if f is None:
            raise ValueError("Filter '%s' not present in this pipeline" % fclass)

        # This will automatically trigger a recompile (at the appropriate level
        # of the hierarchy) if any compile-time parameter is changed.
        #
        f.updateConfigurationFromKeywords(**kwargs)


    def setFilter(self, fclass, **kwargs):
        """Add or reconfigure a filter.

        If a filter of type fclass is not yet enabled, an instance is created and added.
        Any filter parameters not specified will use (filter type specific) default values.

        If a filter of type fclass is already enabled, its configuration is updated.
        Any filter parameters not specified will retain their current values.


        This is one way to enable filters.

        For adding a filter by its type (class) and optionally configuring it at creation time,
        see addFilter().

        For adding a manually created Filter instance, see addFilterInstance().


        Parameters:

          fclass = Filter subclass class object (BlurSharpenFilter, ViewGlowFilter, ...).
                   (NOT an instance object.)

        Kwargs:

          Zero or more filter parameters supported by objects of type fclass.

        Return value:

          The filter instance.

        """
        f = self.findFilter(fclass)
        if f is None:
            # New filter. Add it.
            f = self.addFilter( fclass, **kwargs )
        else:
            # Existing filter. Only change the configuration.
            f.updateConfigurationFromKeywords(**kwargs)

        # Both FilterPipeline.addFilter() and Filter.updateConfigurationFromKeywords()
        # already trigger recompiles if needed; we don't need to do that here.

        return f


    # TODO: rethink case with multiple instances of the same filter
    #
    # TODO: Supporting the mapping of a single filter instance to multiple pipelines
    # TODO: (which would automatically reflect the same configuration in multiple views)
    # TODO: would require at least the following modifications:
    #    - The "mergings" field must be replaced with a dict with one entry per host pipeline.
    #    - attach() and detach() must account for multiple hosts.
    #    - Pipeline-level compile-time parameters must broadcast the change to all hosts.
    #
    def addFilterInstance(self, f):
        """Add a manually created Filter instance to this pipeline.

        A filter of the same class must not already exist in this pipeline.

        The same Filter instance MUST NOT be added to multiple pipelines; if you need
        to have the same filter in multiple views, read the configuration via
        Filter.getConfigurationAsKeywords() and create a new Filter instance
        for the other FilterPipeline with the same configuration options.


        This is one way to enable filters.

        For adding a filter by its type (class) and optionally configuring it at creation time,
        see addFilter().

        For an automatic "add or reconfigure" in the style of the old API, see setFilter().

        """
        f_old = self.findFilter( f.__class__ )
        if f_old is not None:
            raise ValueError("Filter '%s' already present in this pipeline" % f.__class__)

        self.filters.add( f )
        self._needsCompile = True
        self.reconfigure()


    def _createFilterStages(self):
        """Determine and create FilterStages. Part of pipeline rebuild process.

        Intended to be called from reconfigure() after self.cleanup() has been called
        and the filters have been attached to the pipeline. Not intended to support
        other call sites; hence hidden method.

        This maps logical filter stages (Filter.STAGES) to actual render pass FilterStages,
        creates the needed FilterStage instances in self.stages, and assigns the Filter
        instances from self.filters to them in the appropriate manner.

        The mapping is determined dynamically. The needed number of render passes and which
        filters should go into which pass are computed from currently enabled filters
        (specifically, their stageName and isMergeable parameters).

        Stages without any filters enabled will be skipped. Any stages containing only
        mergeable filters will be merged to the end of the previous non-empty render pass
        (multiple successive stages may be concatenated, preserving ordering, to the end
         of the same target).

        The result is that the filters will be assigned into an optimal number of render passes,
        where each render pass consists of filters from one or more logical stages.

        The stageName of each FilterStage will be of the format "[stage1 + stage2 + ... + stageN]",
        where "stageX" are the names of the original logical stages that have been combined
        to produce the render pass.

        If a render pass gets only a single logical stage assigned to it, it will be named "[stage]".


        Return value:

            bool: False if no filters have been enabled in this FilterPipeline, otherwise True.

        """
        # simple struct to make the counting code self-documenting
        class FilterKindCounts:
            def __init__(self, mergeable=0, nonmergeable=0):
                self.mergeable    = mergeable
                self.nonmergeable = nonmergeable

        # Discover the stages needed by currently enabled filters, and count how many
        # mergeable and nonmergeable filters currently want to be placed in each stage.
        #
        # We will use this information to convert the logical stages into render passes,
        # where each render pass may combine multiple successive logical stages.
        # When we do that (later below), each stage (except the first nonempty stage)
        # will be merged into the previous render pass, unless it contains at least
        # one nonmergeable filter.
        #
        filterKindsByStage = {}
        filtersByStage = {}
        for f in self.filters:
            # Sanity check that the stage is known; we need to be able to order the stages correctly
            # in order to set up the pipeline.
            #
            # TODO: if multiple instances of the same filter, need to identify the instance
            # in the error message.
            #
            if f.stageName not in Filter.STAGES:
                raise ValueError("The stageName of %s references undefined stage '%s'. List of valid stages: %s." % (f.__class__.__name__, f.stageName, Filter.STAGES))

            if f.stageName not in filterKindsByStage:
                filterKindsByStage[f.stageName] = FilterKindCounts(mergeable=int(f.isMergeable),
                                                                   nonmergeable=int(not f.isMergeable))
            else:
                filterKindsByStage[f.stageName].mergeable    += int(f.isMergeable)
                filterKindsByStage[f.stageName].nonmergeable += int(not f.isMergeable)

            # Make reverse lookup dict: list of enabled filters keyed by stageName.
            # We will need this when assigning filters to the FilterStages.
            #
            if f.stageName not in filtersByStage:
                filtersByStage[f.stageName] = []
            filtersByStage[f.stageName].append( f )


        # This dict will gather stage name mappings:
        #
        #   - key   = original stageName
        #   - value = target stageName to which the stage has been mapped (by merging)
        #
        stageNameToMerged = {}

        # Find first stage in the global ordering that has at least one enabled filter.
        #
        found = False
        for idx,stageName in enumerate(Filter.STAGES):
            if stageName not in filtersByStage:  # skip empty stages
                continue

            # The first enabled stage obviously cannot be merged to a previous one, so we map it to itself.
            #
            # We also set it as the merge target for successive stages.
            #
            # If any mergings occur, we will assemble the descriptive name of the combined stage
            # into combinedStageName.
            #
            combinedStageName            = stageName
            mergeTargetStageName         = stageName
            stageNameToMerged[stageName] = stageName

            found = True
            break
        if not found:  # no filters enabled; no stages need to be configured.
            return False

        # This will gather human-readable descriptive names, documenting which original
        # stages have been combined into each final stage.
        #
        # We will use these as the stageNames of the actual FilterStages (render passes) we create.
        #
        # This information must be stored in an ordered container, so that we can connect
        # the FilterStages to each other correctly. Thus we use a list.
        #
        finalStageNames = []

        # This maps a finalStageName to its index in finalStageNames.
        # Used during construction of the FilterStages. 
        #
        finalStageNameToIndex = {}

        # This will gather mappings from the original names of the enabled stages to the
        # final human-readable ones.
        #
        # This intermediate step is needed because the final names are assembled dynamically.
        # During the processing, a later stage, if it gets merged, may contribute to the
        # name of the same final stage as the current stage being processed.
        #
        #   - key   = value in stageNameToMerged (this information is available during processing)
        #   - value = one of finalStageNames
        #
        mergedToFinal = {}


        # Determine stage mergings.
        #
        # First, check if we have only one enabled stage.
        #
        # If the first enabled stage is the last stage defined in Filter.STAGES, we will have only one stage.
        # Otherwise, check if at least one remaining stage is nonempty.
        #
        singleStage = (idx == len(Filter.STAGES) - 1)
        if not singleStage:
            found = False
            for stageName in Filter.STAGES[idx+1:]:
                if stageName in filtersByStage:
                    found = True  # this stage is not empty
                    break
            singleStage = not found

        if not singleStage:
            # At this point, we have at least one more non-empty stage remaining.
            #
            # Walk the remaining stages and set up merges.
            #
            lastProcessed = ""  # needed for finalization of last render pass
            for stageName in Filter.STAGES[idx+1:]:
                # Skip empty stages.
                #
                # It doesn't matter that empty stages are not counted in the number of mergings,
                # because the purpose of counting mergings is only to preserve ordering.
                # 
                if stageName not in filtersByStage:
                    continue

                # If no nonmergeable filters enabled in this stage,
                # merge this stage into the current merge target stage.
                #
                if filterKindsByStage[stageName].nonmergeable == 0:
                    combinedStageName += " + %s" % stageName
                    stageNameToMerged[stageName] = mergeTargetStageName

                    # Update number of mergings for filters in this stage to make sort preserve ordering
                    # even though several logical stages end up in the same render pass (and thus in
                    # the same FilterStage).
                    #
                    for f in filtersByStage[stageName]:
                        f.mergings += 1

                    lastProcessed = "mergeable"

                else:  # at least one nonmergeable filter in this stage
                    # Now that we found a stage that cannot be merged,
                    # the old combined stage name is final.
                    #
                    # Add it to final stages; a render pass will be created for it.
                    #
                    stylizedName = "[%s]" % combinedStageName
                    finalStageNames.append( stylizedName )
                    finalStageNameToIndex[stylizedName] = len(finalStageNames) - 1
                    mergedToFinal[mergeTargetStageName] = stylizedName

                    # Make this stage the new merge target and begin constructing a new render pass.
                    #
                    combinedStageName            = stageName
                    mergeTargetStageName         = stageName
                    stageNameToMerged[stageName] = stageName

                    lastProcessed = "nonmergeable"

        # Finalize the last (or only) render pass if necessary; only the case "nonmergeable"
        # has already updated this information.
        #
        if singleStage  or  lastProcessed == "mergeable":
            stylizedName = "[%s]" % combinedStageName
            finalStageNames.append( stylizedName )
            finalStageNameToIndex[stylizedName] = len(finalStageNames) - 1
            mergedToFinal[mergeTargetStageName] = stylizedName

        # At this point:
        #   - We have at least one render pass.
        #   - finalStageNames contains an ordered list of final stage names (render pass names),
        #     accounting for any merged stages.
        #   - Each original stage name maps to one of the final stage names (render pass names)
        #     as "mergedToFinal[ stageNameToMerged[stageName] ]".

        # Create the FilterStages representing the render passes.
        #
        for s in finalStageNames:
            stage = FilterStage( pipeline=self, stageName=s )

            # We must handle StageInitializationFilter specially, because
            # it must not participate in the logical stage merging process.
            #
            # One initialization filter must be inserted in each actual render pass.
            #
            # Other filters have already been attached to the pipeline when _createFilterStages()
            # is called. We must do that also for the initialization filter to make it run its setup()
            # (which registers the input texture to the pipeline).
            #
            initFilter = StageInitializationFilter()
            initFilter.attach( pipeline=self )
            stage.addFilterInstance( initFilter )

            self.stages.append( stage )

        # Assign our Filters to the created FilterStages.
        #
        for f in self.filters:
            s = mergedToFinal[ stageNameToMerged[f.stageName] ]  # final stage name
            stage = self.stages[ finalStageNameToIndex[s] ]
            stage.addFilterInstance( f )

        return True


    def _propagateRegistrationData(self):
        """Propagate any relevant registration data to FilterStages.

        Internal method, called by reconfigure() after FilterStages have been created.

        """
        for stage in self.stages:
            # See if the filters assigned to this stage registered anything to us,
            # and pass the relevant information to the FilterStage.
            for f in stage.filters:
                if f in self.registeredTextures:
                    for texName,needTexpix in self.registeredTextures[f]:
                        stage._registerInputTexture(f, texName, needTexpix)

                if f in self.registeredCustomInputs:
                    for inputType,inputName in self.registeredCustomInputs[f]:
                        stage._registerCustomInput(f, inputType, inputName)

                if f in self.updateTasks:
                    stage._registerUpdatable(f, self.updateTasks[f])


    def reconfigure(self, force=False):
        """Rebuild the pipeline.

        If no filters are enabled, causes the screen to be tinted pink. This behavior
        comes from FilterManager, and is intended as a visual clue that an unnecessary,
        blank filter pipeline is currently enabled.

        If any filters are enabled, sets up postprocessing using those filters.

        Note that the filters retain their configuration across reconfigure() calls.

        """

        # During a batch update to filter configuration, do not recompile.
        #
        if self.batchUpdateInProgress:
            return

        # Unless forced, compile only if needed.
        #
        if not force  and  not self._needsCompile:
            return

        # Detach filters from FilterStages and destroy the old FilterStages.
        # This also detaches the filters from the pipeline.
        #
        self.cleanup()


        ## We could do stuff here if we needed to do something specifically
        ## at the time when the filters are not attached to the pipeline.


        # Attach the filters to the pipeline.
        #
        # This calls setup() for the pipeline phase, registering input textures, custom shader inputs
        # and aux bits.
        #
        # FilterPipeline needs the (consolidated) input textures and aux bits;
        # the rest gets passed on to FilterStage when we create those in _createFilterStages().
        #
        # This must be done *before* creating the stages, so that _createFilterStages() has access
        # to the registered input texture and custom shader input information.
        #
        try:
            for f in self.filters:
                f.attach(pipeline=self)
        except:
            self.cleanup()
            raise


        # Map logical stages to render passes (FilterStage instances) based on which filters are enabled.
        # Create the FilterStage instances, and assign the filters to them.
        #
        # This also creates StageInitializationFilters, one for each stage. They are special:
        # they have no meaningful properties, and are owned by the FilterStage.
        #
        # (FilterPipeline would not know what to do with multiple instances of the same Filter subclass.
        #  Also, the initialization filters are not intended to be visible to the user, and they should
        #  get destroyed automatically when the FilterStages are destroyed during the next cleanup().)
        #
        try:
            ret = self._createFilterStages()
        except:
            self.cleanup()
            raise
        if ret == False:
            # We have no filters enabled, except maybe...
            #
            # HalfPixelShift is not an fshader function like the other filters, so it has no corresponding
            # Filter object to represent it. Instead, it is an option that modifies the generated vshader
            # of a FilterStage.
            #
            # If the user wants to enable only HalfPixelShift, we need to generate a stage that has
            # just the initialization filter enabled, but with the HalfPixelShift flag set.
            #
            if self.halfPixelShift:
                # We perform the official stage setup procedure (the same steps _createFilterStages() does)
                # to avoid mistakes.

                # stageName doesn't need to be one of Filter.STAGES, because only
                # _createFilterStages() requires that, and only for f.stageName,
                # where f is a *Filter*. (Observe that here we are naming a *FilterStage*.)
                #
                stage = FilterStage( pipeline=self, stageName="[HalfPixelShiftOnly]" )

                initFilter = StageInitializationFilter()
                initFilter.attach( pipeline=self )  # make it register the color texture to the pipeline
                stage.addFilterInstance( initFilter )

                self.stages.append( stage )

                # Now we have one stage, which has one enabled filter, and we can continue with the
                # usual reconfigure process (which will set the stage's HalfPixelShift flag later).
            else:
                # No filters enabled (not even HalfPixelShift).
                # Let FilterManager set default pink tint and return.
                #
                # TODO: make the default a passthrough instead?
                #  * pink tint pros/cons:
                #     + cannot accidentally forget FilterPipeline enabled when not needed
                #     - if filters can be switched on and off in an interactive application,
                #       the behavior when the last filter is disabled is unintuitive.
                #  * passthrough pros/cons:
                #     * The opposite of those of pink tint.
                #
                self.sceneTextures["color"] = makeFilterTexture("color")
                finalQuad = self.manager.renderSceneInto(textures=self.sceneTextures, auxbits=0)
                if finalQuad is None:
                    self.cleanup()
                    raise RuntimeError("FilterManager.renderSceneInto() failed during setting default pink tint.")
                self.quads.append( finalQuad )
                return  # setup success (although no filters enabled)


        # FilterStages have now been created.
        #
        # At this point, we have at least one FilterStage.
        #
        assert( len(self.stages) >= 1 )

        # Propagate the registrations (input textures, custom shader inputs, update tasks)
        # provided by the Filters to the relevant FilterStages.
        #
        self._propagateRegistrationData()

        # Create textures and quads. Connect stages via FilterManager.
        #
        # First, see which scene textures we will need. We only create those scene textures
        # that are actually needed by at least one enabled filter (including the initialization filter).
        #
        # We ignore any internal textures defined by the filters, as they are created by the filters themselves.
        #
        neededSceneTextures = filter( lambda s: s in Filter.SCENETEXTURES, self.consolidatedTextures.keys() ) 
        for texName in neededSceneTextures:
            self.sceneTextures[texName] = makeFilterTexture(texName)

        # Note that FilterManager.renderSceneInto() returns the *final* quad, which is actually shown on screen,
        # although rendering the scene into a set of textures is the *first* step that occurs in the pipeline
        # (before any filter stages proper).
        #
        # The shader that this last quad expects is that which will render the final output, i.e. the shader of
        # the *final* stage in the pipeline.
        #
        finalQuad = self.manager.renderSceneInto(textures=self.sceneTextures, auxbits=self.consolidatedAuxBits)
        if finalQuad is None:
            self.cleanup()
            raise RuntimeError("FilterManager.renderSceneInto() failed during setup of finalQuad. Aux bits = %s, scene textures = %s." % (self.consolidatedAuxBits, self.sceneTextures))

        # All stages except the final one are intermediate stages, which render into a texture.
        #
        # output from scene = input to first stage
        self.stageColorTextures.append( self.sceneTextures["color"] )
        for i,stage in enumerate(self.stages[:-1]):
            outputTexture = makeFilterTexture("stage%d" % i)  # output texture for this stage
            outputQuad    = self.manager.renderQuadInto( colortex=outputTexture )
            if outputQuad is None:
                self.cleanup()
                raise RuntimeError("FilterManager.renderQuadInto() failed during setup of stage '%s' (%d out of %d stages). Aux bits = %s, scene textures = %s." % (stage.stageName, i+1, len(self.stages), self.consolidatedAuxBits, self.sceneTextures))

            # We pass through "depth" and "aux" from the scene.
            #
            # TODO: Support updating aux and depth in the filters, via additional output textures
            #       (auxiliary render targets). Without that support, depth and aux are not usable
            #       after LensDistortion or similar image remapping (deforming) filters.
            #
            stageInputTextures = self.sceneTextures.copy()  # shallow-copy references only
            stageInputTextures["color"] = self.stageColorTextures[-1]  # input color = previous output

            # Assign the input textures to the FilterStage, and select the quad
            # into which the compositing shader (from this FilterStage) renders.
            #
            stage.connect( stageInputTextures, outputQuad )

            self.stageColorTextures.append( outputTexture )
            self.quads.append( outputQuad )

        # The final stage renders into finalQuad, which is shown on screen.
        # It does not need an output texture.
        #
        stageInputTextures = self.sceneTextures.copy()  # shallow-copy references only
        stageInputTextures["color"] = self.stageColorTextures[-1]  # previous output
        self.stages[-1].connect( stageInputTextures, finalQuad )
        self.quads.append( finalQuad )

        # Pass through the current value of our halfPixelShift setting to the first FilterStage.
        #
        # We write to the internal variable in FilterStage directly, because we don't want
        # the setter to call reconfigure() here.
        #
        self.stages[0]._halfPixelShift = self.halfPixelShift

        # Texture and quad setup complete.
        #
        # The final step: compile all created FilterStages.
        #
        # When FilterStage attaches its assigned Filters to the stage, each Filter will compile itself,
        # and get references to the currently connected stage input textures from FilterStage.
        #
        # Clearing the allowDelegation flag here tells FilterStage that FilterPipeline is calling it;
        # i.e. the call did not come from a property setter. This prevents FilterStage from escalating
        # the call back to us (which would cause infinite recursion). We need this mechanism because
        # recompiling a single stage is not possible (grep tag: FilterManager, monolithic cleanup),
        # so also single-stage recompiles are done by delegating the work to FilterPipeline,
        # so that the whole pipeline is rebuilt.
        #
        try:
            for stage in self.stages:
                stage.reconfigure(allowDelegation=False)
        except:
            self.cleanup()
            raise

        # The rebuild is now complete.
        #
        self._needsCompile = False


########################################
# Backward compatibility layer
########################################

class CommonFilters(FilterPipeline):
    """This class adds a backward-compatible API (Panda 1.8.x) on top of the FilterPipeline API (introduced in 1.9.0)."""

    def __init__(self, win, cam):
        """Constructor.

        win and cam like in FilterManager; see also FilterPipeline, which uses the same parameters.

        """
        super(self.__class__, self).__init__(win, cam)

    def _compatSetFilter(fclass, **kwargs):
        """Backward-compatible backend for the old API set***Filter() methods.

        This is a thin wrapper for setFilter(), which simply discards exceptions,
        returning True if successful and False on failure (caught an exception).

        Do not use this except for backward compatibility purposes. If something
        goes wrong, setFilter() is much better, because all the exceptions raised
        by the postprocessing filter subsystem contain an error message that is
        packed with as much information as possible at the site where the exception
        was raised.

        """
        # Note:
        #
        # This sends a christmas tree to the filter, because in the old API, the new configuration
        # always completely replaces the previous one - hence we are always called with a full set
        # of kwargs for the filter being configured (with defaults for any non-specified ones,
        # coming from the default argument values of the setThisAndThatFilter() methods;
        # these are not necessarily the default values set in the Filter implementations).
        #
        # We don't need to worry about that, though, because compile-time properties created
        # using makeCTParam() are intelligent enough to trigger a recompile only when the
        # new value of the parameter is actually different from its previous value.
        #
        try:
            self.setFilter(fclass, **kwargs)
        except:
            return False
        return True

    #######################
    # Filter set/del stubs
    #######################

    def setHalfPixelShift(self):
        # TODO docstring here
        try:
            self.halfPixelShift = True  # property, runs setter, which runs reconfigure
        except:
            return False
        return True

    def delHalfPixelShift(self):
        self.halfPixelShift = False
        return True

    def setBlurSharpen(self, amount=0.0):
        # TODO docstring here
        return self._compatSetFilter(BlurSharpenFilter, amount=amount)

    def delBlurSharpen(self):
        self.delFilter(BlurSharpenFilter)
        return True

    def setViewGlow(self):
        # TODO docstring here
        return self._compatSetFilter(ViewGlowFilter)

    def delViewGlow(self):
        self.delFilter(ViewGlowFilter)
        return True

    # etc.

    # TODO: The following filters didn't exist in the old CommonFilters; what to do?
    #
    # Options:
    #
    #  a) The old API was pretty good for simple use cases, where the new one would feel overly bureaucratic.
    #     Continue supporting the old API as an alternative. Add old-style API support for all new filters.
    #
    #  b) Mark the old API as deprecated. Encourage developers to write new scripts using the new one.
    #     However, add old-style API support for all new filters, to facilitate easily using them
    #     in old scripts/applications with only very minor modifications to existing code.
    #
    #  c) Mark the old API as deprecated. Make the new version backward compatible, but no more.
    #     Leave any new filters exclusive to the new API.
    #
    def setTint(self, mode="fade", color=(1,1,1,1), blend=1.0):
        # TODO docstring here
        return self._compatSetFilter(TintFilter, mode, color, blend)

    def delTint(self):
        self.delFilter(TintFilter)
        return True

    def setScanlines(self, blend=0.5, thickness=1, dynamic=False, top=True):
        # TODO docstring here
        # TODO: parameter "dynamic": dynamic scanlines are not yet supported (needs update task)
        field = ScanlinesFilter.FIELD_TOP  if top == True  else  ScanlinesFilter.FIELD_BOTTOM
        return self._compatSetFilter(ScanlinesFilter, blend=blend, line_thickness=thickness, field=field)

    def delScanlines(self):
        self.delFilter(ScanlinesFilter)
        return True

